import logging
import os
import plistlib

import click
import coloredlogs

NOTIFICATIONS_FILENAME = os.path.join(os.path.dirname(__file__), "notifications.txt")


def get_notifications():
    with open(NOTIFICATIONS_FILENAME, "rb") as f:
        return f.read().decode().split("\n")


def save_notifications(notifications: list[str]):
    with open(NOTIFICATIONS_FILENAME, "wb") as f:
        notifications.sort()
        f.write("\n".join(notifications).encode())


@click.command()
@click.argument("root_fs", type=click.Path(dir_okay=True, file_okay=False, exists=True))
def main(root_fs):
    """
    Add notifications registered to `com.apple.notifyd.matching` from a given IPSW `root_fs` (extracted filesystem)
    into `notifications.txt`
    """
    launch_daemons = os.path.join(root_fs, "System", "Library", "LaunchDaemons")

    notifications = set(get_notifications())

    for filename in os.listdir(launch_daemons):
        if not filename.endswith(".plist"):
            continue

        filename = os.path.join(launch_daemons, filename)
        try:
            with open(filename, "rb") as f:
                plist = plistlib.load(f)
        except Exception:
            logging.exception(f"error parsing: {filename}")
            continue

        launch_events = plist.get("LaunchEvents", {})
        notifyd_matching = launch_events.get("com.apple.notifyd.matching", {})

        for v in notifyd_matching.values():
            if not isinstance(v, dict):
                logging.error(f"error parsing: {filename}")
                continue
            notification = v.get("Notification")
            if notification is None:
                continue

            if notification not in notifications:
                logging.info(f"adding notification: {notification}")
                notifications.add(notification)

    save_notifications(list(notifications))


if __name__ == "__main__":
    coloredlogs.install(level=logging.DEBUG)
    main()



import asyncio
import logging
import random
from abc import abstractmethod

import gpxpy


class LocationSimulationBase:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @abstractmethod
    async def set(self, latitude: float, longitude: float) -> None:
        pass

    @abstractmethod
    async def clear(self) -> None:
        pass

    async def play_gpx_file(self, filename: str, disable_sleep: bool = False, timing_randomness_range: int = 0):
        with open(filename) as f:
            gpx = gpxpy.parse(f)

        last_time = None
        gpx_timing_noise = None
        for track in gpx.tracks:
            for segment in track.segments:
                for point in segment.points:
                    if last_time is not None:
                        duration = (point.time - last_time).total_seconds()
                        if duration >= 0 and not disable_sleep:
                            if timing_randomness_range:
                                gpx_timing_noise = (
                                    random.randint(-timing_randomness_range, timing_randomness_range) / 1000
                                )
                                duration += gpx_timing_noise

                            self.logger.info(f"waiting for {duration:.3f}s")
                            await asyncio.sleep(duration)
                    last_time = point.time
                    self.logger.info(f"set location to {point.latitude} {point.longitude}")
                    await self.set(point.latitude, point.longitude)



lpol_file = bytes([
    0x30,
    0x14,
    0x16,
    0x04,
    0x49,
    0x4D,
    0x34,
    0x50,
    0x16,
    0x04,
    0x6C,
    0x70,
    0x6F,
    0x6C,
    0x16,
    0x03,
    0x31,
    0x2E,
    0x30,
    0x04,
    0x01,
    0x00,
])

# extracted from ac2
PROGRESS_BAR_OPERATIONS = {
    11: "CREATE_PARTITION_MAP",
    12: "CREATE_FILESYSTEM",
    13: "RESTORE_IMAGE",
    14: "VERIFY_RESTORE",
    15: "CHECK_FILESYSTEMS",
    16: "MOUNT_FILESYSTEMS",
    17: "FIXUP_VAR",
    18: "FLASH_FIRMWARE",
    19: "UPDATE_BASEBAND",
    20: "SET_BOOT_STAGE",
    21: "REBOOT_DEVICE",
    22: "SHUTDOWN_DEVICE",
    23: "TURN_ON_ACCESSORY_POWER",
    24: "CLEAR_BOOTARGS",
    25: "MODIFY_BOOTARGS",
    26: "INSTALL_ROOT",
    27: "INSTALL_KERNELCACHE",
    28: "WAIT_FOR_NAND",
    29: "UNMOUNT_FILESYSTEMS",
    30: "SET_DATETIME",
    31: "EXEC_IBOOT",
    32: "FINALIZE_NAND_EPOCH_UPDATE",
    33: "CHECK_INAPPR_BOOT_PARTITIONS",
    34: "CREATE_FACTORY_RESTORE_MARKER",
    35: "LOAD_FIRMWARE",
    36: "REQUESTING_FUD_DATA",
    37: "REMOVING_ACTIVATION_RECORD",
    38: "CHECK_BATTERY_VOLTAGE",
    39: "WAIT_BATTERY_CHARGE",
    40: "CLOSE_MODEM_TICKETS",
    41: "MIGRATE_DATA",
    42: "WIPE_STORAGE_DEVICE",
    43: "SEND_APPLE_LOGO",
    44: "CHECK_LOGS",
    46: "CLEAR_NVRAM",
    47: "UPDATE_GAS_GAUGE",
    48: "PREPARE_BASEBAND_UPDATE",
    49: "BOOT_BASEBAND",
    50: "CREATE_SYSTEM_KEYBAG",
    51: "UPDATE_IR_MCU_FIRMWARE",
    52: "RESIZE_SYSTEM_PARTITION",
    53: "COLLECTING_UPDATER_OUTPUT",
    54: "PAIR_STOCKHOLM",
    55: "UPDATE_STOCKHOLM",
    56: "UPDATE_SWDHID",
    57: "CERTIFY_SEP",
    58: "UPDATE_NAND_FIRMWARE",
    59: "UPDATE_SE_FIRMWARE",
    60: "UPDATE_SAVAGE",
    61: "INSTALLING_DEVICETREE",
    62: "CERTIFY_SAVAGE",
    63: "SUBMITTING_PROVINFO",
    64: "CERTIFY_YONKERS",
    65: "UPDATE_ROSE",
    66: "UPDATE_VERIDIAN",
    67: "CREATING_PROTECTED_VOLUME",
    68: "RESIZING_MAIN_FS_PARTITION",
    69: "CREATING_RECOVERY_OS_VOLUME",
    70: "INSTALLING_RECOVERY_OS_FILES",
    71: "INSTALLING_RECOVERY_OS_IMAGE",
    74: "REQUESTING_EAN_DATA",
    77: "SEALING_SYSTEM_VOLUME",
    81: "UPDATING_APPLETCON",
}



import socket
from collections.abc import AsyncGenerator
from typing import Optional, Union

from pymobiledevice3.exceptions import NotificationTimeoutError
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.lockdown_service import LockdownService


class NotificationProxyService(LockdownService):
    SERVICE_NAME = "com.apple.mobile.notification_proxy"
    RSD_SERVICE_NAME = "com.apple.mobile.notification_proxy.shim.remote"

    INSECURE_SERVICE_NAME = "com.apple.mobile.insecure_notification_proxy"
    RSD_INSECURE_SERVICE_NAME = "com.apple.mobile.insecure_notification_proxy.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider, insecure=False, timeout: Optional[Union[float, int]] = None):
        if isinstance(lockdown, RemoteServiceDiscoveryService):
            secure_service_name = self.RSD_SERVICE_NAME
            insecure_service_name = self.RSD_INSECURE_SERVICE_NAME
        else:
            secure_service_name = self.SERVICE_NAME
            insecure_service_name = self.INSECURE_SERVICE_NAME

        if insecure:
            super().__init__(lockdown, insecure_service_name)
        else:
            super().__init__(lockdown, secure_service_name)

        if timeout is not None:
            self.service.socket.settimeout(timeout)

    async def notify_post(self, name: str) -> None:
        """Send notification to the device's notification_proxy."""
        await self.service.send_plist({"Command": "PostNotification", "Name": name})

    async def notify_register_dispatch(self, name: str) -> None:
        """Tells the device to send a notification on the specified event."""
        self.logger.info(f"Observing {name}")
        await self.service.send_plist({"Command": "ObserveNotification", "Name": name})

    async def receive_notification(self) -> AsyncGenerator[dict, None]:
        while True:
            try:
                yield await self.service.recv_plist()
            except socket.timeout as e:
                raise NotificationTimeoutError from e



from packaging.version import Version

from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.remote_server import RemoteServer


class DvtSecureSocketProxyService(RemoteServer):
    SERVICE_NAME = "com.apple.instruments.remoteserver.DVTSecureSocketProxy"
    OLD_SERVICE_NAME = "com.apple.instruments.remoteserver"
    RSD_SERVICE_NAME = "com.apple.instruments.dtservicehub"

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, RemoteServiceDiscoveryService):
            service_name = self.RSD_SERVICE_NAME
            remove_ssl_context = False
        elif Version(lockdown.product_version) >= Version("14.0"):
            service_name = self.SERVICE_NAME
            remove_ssl_context = False
        else:
            service_name = self.OLD_SERVICE_NAME
            remove_ssl_context = True

        super().__init__(lockdown, service_name, remove_ssl_context=remove_ssl_context)



from pymobiledevice3.services.dvt.instruments import ChannelService
from pymobiledevice3.services.remote_server import MessageAux


class EnergyMonitor(ChannelService):
    IDENTIFIER = "com.apple.xcode.debug-gauge-data-providers.Energy"

    def __init__(self, dvt, pid_list: list) -> None:
        super().__init__(dvt)
        self._pid_list = pid_list

    async def __aenter__(self):
        channel = await self._channel_ref()
        # stop monitoring if already monitored
        await channel.stopSamplingForPIDs_(MessageAux().append_obj(self._pid_list))

        await channel.startSamplingForPIDs_(MessageAux().append_obj(self._pid_list))
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        channel = await self._channel_ref()
        await channel.stopSamplingForPIDs_(MessageAux().append_obj(self._pid_list))

    def __aiter__(self):
        while True:
            yield self._sample_once()

    async def _sample_once(self):
        channel = await self._channel_ref()
        await channel.sampleAttributes_forPIDs_(MessageAux().append_obj({}).append_obj(self._pid_list))
        return await channel.receive_plist()



#!/usr/bin/env python3
import dataclasses
import logging
import plistlib
import xml.etree.ElementTree as ET
from contextlib import aclosing
from pathlib import Path
from typing import Optional

import click
import inquirer3
import requests
from requests.structures import CaseInsensitiveDict

from pymobiledevice3.exceptions import MobileActivationException
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider

ACTIVATION_USER_AGENT_IOS = "iOS Device Activator (MobileActivation-20 built on Jan 15 2012 at 19:07:28)"
ACTIVATION_DEFAULT_URL = "https://albert.apple.com/deviceservices/deviceActivation"
ACTIVATION_DRM_HANDSHAKE_DEFAULT_URL = "https://albert.apple.com/deviceservices/drmHandshake"
DEFAULT_HEADERS = {
    "Accept": "application/xml",
    "User-Agent": ACTIVATION_USER_AGENT_IOS,
    "Expect": "100-continue",
}

ACTIVATION_REQUESTS_SUBDIR = Path("offline_requests")
NONCE_CYCLE_INTERVAL = 60 * 5


@dataclasses.dataclass
class Field:
    id: str
    label: str
    placeholder: str
    secure: bool


@dataclasses.dataclass
class ActivationForm:
    title: str
    description: str
    fields: list[Field]
    server_info: dict[str, str]


class MobileActivationService:
    """
    Perform device activation

    There is no point in inheriting from BaseService since we'll need a new lockdown connection
    for each request.
    """

    SERVICE_NAME = "com.apple.mobileactivationd"

    def __init__(self, lockdown: LockdownServiceProvider) -> None:
        self.lockdown = lockdown
        self.logger = logging.getLogger(__name__)

    async def state(self):
        try:
            return (await self.send_command("GetActivationStateRequest"))["Value"]
        except Exception:
            return await self.lockdown.get_value(key="ActivationState")

    async def wait_for_activation_session(self):
        try:
            blob = await self.create_activation_session_info()
        except Exception:
            return
        handshake_request_message = blob["HandshakeRequestMessage"]
        while handshake_request_message == blob["HandshakeRequestMessage"]:
            blob = await self.create_activation_session_info()

    @staticmethod
    def _get_activation_form_from_response(content: str) -> ActivationForm:
        root = ET.fromstring(content)
        title = root.find("page/navigationBar").get("title")
        description = root.find("page/tableView/section/footer").text
        fields = []
        for editable in root.findall("page//editableTextRow"):
            fields.append(
                Field(
                    id=editable.get("id"),
                    label=editable.get("label"),
                    placeholder=editable.get("placeholder"),
                    secure=bool(editable.get("secure", False)),
                )
            )
        server_info = {}
        for k, v in root.find("serverInfo").items():
            server_info[k] = v
        return ActivationForm(title=title, description=description, fields=fields, server_info=server_info)

    async def activate(self, skip_apple_id_query: bool = False) -> None:
        if await self.state() != "Unactivated":
            self.logger.error("Device is already activated!")
            return

        try:
            blob = await self.create_activation_session_info()
            session_mode = True
        except Exception:
            session_mode = False

        # create drmHandshake request with blob from device
        headers = {"Content-Type": "application/x-apple-plist"}
        headers.update(DEFAULT_HEADERS)
        if session_mode:
            content, headers = self.post(
                ACTIVATION_DRM_HANDSHAKE_DEFAULT_URL, data=plistlib.dumps(blob), headers=headers
            )

        activation_request = {}
        if session_mode:
            activation_info = await self.create_activation_info_with_session(content)
        else:
            activation_info = await self.lockdown.get_value(key="ActivationInfo")
            activation_request.update({
                "InStoreActivation": False,
                "AppleSerialNumber": await self.lockdown.get_value(key="SerialNumber"),
            })
            if self.lockdown.all_values.get("TelephonyCapability"):
                req_pair = {
                    "IMEI": "InternationalMobileEquipmentIdentity",
                    "MEID": "MobileEquipmentIdentifier",
                    "IMSI": "InternationalMobileSubscriberIdentity",
                    "ICCID": "IntegratedCircuitCardIdentity",
                }

                has_meid = False
                for k, v in req_pair.items():
                    lv = self.lockdown.all_values.get(v)
                    if lv is not None:
                        activation_request.update({k: lv})
                        continue
                    else:
                        self.logger.warn(f"Unable to get {k} from lockdownd")
                        if k == "MEID" and has_meid:
                            # Something is wrong if both IMEI & MEID is missing
                            raise MobileActivationException("Unable to obtain both IMEI and MEID")

                    # Either IMEI or MEID, or both
                    if k == "IMEI":
                        has_meid = lv is None
        activation_request.update({"activation-info": plistlib.dumps(activation_info)})

        content, headers = self.post(ACTIVATION_DEFAULT_URL, data=activation_request)
        content_type = headers["Content-Type"]

        if content_type == "application/x-buddyml":
            if skip_apple_id_query:
                raise MobileActivationException("Device is iCloud locked")
            try:
                activation_form = self._get_activation_form_from_response(content.decode())
            except Exception as e:
                raise MobileActivationException("Activation server response is invalid") from e
            else:
                click.secho(activation_form.title, bold=True)
                click.secho(activation_form.description)
                fields = []
                for field in activation_form.fields:
                    if field.secure:
                        fields.append(inquirer3.Password(name=field.id, message=f"{field.label}"))
                    else:
                        fields.append(inquirer3.Text(name=field.id, message=f"{field.label}"))
                data = inquirer3.prompt(fields)
                data.update(activation_form.server_info)
                content, headers = self.post(ACTIVATION_DEFAULT_URL, data=data)
                content_type = headers["Content-Type"]

        assert content_type == "text/xml"
        if session_mode:
            await self.activate_with_session(content, headers)
        else:
            await self.activate_with_lockdown(content)

        # set ActivationStateAcknowledged if we succeeded
        await self.lockdown.set_value(True, key="ActivationStateAcknowledged")

    async def deactivate(self):
        try:
            return await self.send_command("DeactivateRequest")
        except Exception:
            return await self.lockdown._request_async("Deactivate")

    async def create_activation_session_info(self):
        response = await self.send_command("CreateTunnel1SessionInfoRequest")
        error = response.get("Error")
        if error is not None:
            raise MobileActivationException(f"Mobile activation can not be done due to: {response}")
        return response["Value"]

    async def create_activation_info_with_session(self, handshake_response):
        response = await self.send_command("CreateTunnel1ActivationInfoRequest", handshake_response)
        error = response.get("Error")
        if error is not None:
            raise MobileActivationException(f"Mobile activation can not be done due to: {response}")
        return response["Value"]

    async def activate_with_lockdown(self, activation_record):
        record = plistlib.loads(activation_record)
        node = record.get("iphone-activation")
        if node is None:
            node = record.get("device-activation")
        if node is None:
            raise MobileActivationException("Activation record received is invalid")

        await self.lockdown._request_async("Activate", {"ActivationRecord": node.get("activation-record")})

    async def activate_with_session(self, activation_record, headers):
        data = {
            "Command": "HandleActivationInfoWithSessionRequest",
            "Value": activation_record,
        }
        if headers:
            data["ActivationResponseHeaders"] = dict(headers)
        async with aclosing(await self.lockdown.start_lockdown_service(self.SERVICE_NAME)) as service:
            return await service.send_recv_plist(data)

    async def send_command(self, command: str, value: str = ""):
        data = {"Command": command}
        if value:
            data["Value"] = value
        async with aclosing(await self.lockdown.start_lockdown_service(self.SERVICE_NAME)) as service:
            return await service.send_recv_plist(data)

    def post(
        self, url: str, data: dict, headers: Optional[CaseInsensitiveDict[str, str]] = None
    ) -> tuple[bytes, CaseInsensitiveDict[str]]:
        if headers is None:
            headers = DEFAULT_HEADERS

        resp = requests.post(url, data=data, headers=headers)
        return resp.content, resp.headers



from construct import Array, Bytes, Const, Default, Int32ub, Int32ul, Pointer, Struct, this

ftab_entry = Struct(
    "tag" / Bytes(4),
    "offset" / Int32ul,
    "size" / Int32ul,
    "pad_0x0C" / Default(Int32ul, 0),
    "data" / Pointer(this.offset, Bytes(this.size)),
)

ftab_header = Struct(
    "always_01" / Int32ul,  # 1
    "always_ff" / Int32ul,  # 0xFFFFFFFF
    "unk_0x08" / Int32ub,  # 0
    "unk_0x0C" / Int32ub,  # 0
    "unk_0x10" / Int32ub,  # 0
    "unk_0x14" / Int32ub,  # 0
    "unk_0x18" / Int32ub,  # 0
    "unk_0x1C" / Int32ub,  # 0
    "tag" / Bytes(4),  # e.g. 'rkos'
    "magic" / Const(b"ftab"),  # 'ftab' magic
    "num_entries" / Int32ul,
    "pad_0x2C" / Int32ub,
    "entries" / Array(this.num_entries, ftab_entry),
)


class Ftab:
    def __init__(self, component_data: bytes):
        self.parsed = ftab_header.parse(component_data)

    @property
    def tag(self):
        return self.parsed.tag

    def get_entry_data(self, tag: bytes) -> bytes:
        for entry in self.parsed.entries:
            if entry.tag == tag:
                return entry.data
        return None

    def add_entry(self, tag: bytes, data: bytes):
        new_offset = self.parsed.entries[-1].offset + self.parsed.entries[-1].size
        new_entry = {"tag": tag, "offset": new_offset, "size": len(data), "data": data}

        self.parsed.num_entries += 1
        self.parsed.entries.append(new_entry)

    @property
    def data(self) -> bytes:
        return ftab_header.build(self.parsed)



import logging

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.services.dvt.instruments import ChannelService
from pymobiledevice3.services.remote_server import MessageAux


class ConditionInducer(ChannelService):
    IDENTIFIER = "com.apple.instruments.server.services.ConditionInducer"

    def __init__(self, dvt):
        self.logger = logging.getLogger(__name__)
        super().__init__(dvt)

    async def list(self) -> list:
        channel = await self._channel_ref()
        await channel.availableConditionInducers()
        return await channel.receive_plist()

    async def set(self, profile_identifier):
        channel = await self._channel_ref()
        for group in await self.list():
            for profile in group.get("profiles"):
                if profile_identifier == profile.get("identifier"):
                    self.logger.info(profile.get("description"))
                    await channel.enableConditionWithIdentifier_profileIdentifier_(
                        MessageAux().append_obj(group.get("identifier")).append_obj(profile.get("identifier"))
                    )
                    # wait for response which may be a raised NSError
                    await channel.receive_plist()
                    return
        raise PyMobileDevice3Exception("Invalid profile identifier")

    async def clear(self):
        channel = await self._channel_ref()
        await channel.disableActiveCondition()



from pathlib import Path
from textwrap import dedent
from typing import Annotated

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command
from pymobiledevice3.cli.developer import (
    accessibility,
    arbitration,
    condition,
    core_device,
    debugserver,
    dvt,
    fetch_symbols,
    simulate_location,
    wda,
)
from pymobiledevice3.services.remote_server import RemoteServer
from pymobiledevice3.services.screenshot import ScreenshotService

cli = InjectingTyper(
    name="developer",
    help=dedent("""\
        Developer tooling for iOS devices (requires Developer Mode + mounted DeveloperDiskImage).

        These commands require the DeveloperDiskImage.dmg to be mounted on the device prior
        to execution. You can achieve this using:

        pymobiledevice3 mounter mount

        Also, starting at iOS 17.0, a tunnel must be created to the device for the services
        to be accessible. Therefore, every CLI command is retried with a `--tunnel` option
        for implicitly accessing tunneld when necessary
    """),
    no_args_is_help=True,
)
cli.add_typer(dvt.cli)
cli.add_typer(fetch_symbols.cli)
cli.add_typer(simulate_location.cli)
cli.add_typer(accessibility.cli)
cli.add_typer(condition.cli)
cli.add_typer(debugserver.cli)
cli.add_typer(arbitration.cli)
cli.add_typer(core_device.cli)
cli.add_typer(wda.cli)


@cli.command("shell")
def developer_shell(
    service_provider: ServiceProviderDep,
    service: str,
    remove_ssl_context: Annotated[bool, typer.Option("--remove-ssl-context", "-r")] = False,
) -> None:
    """Open an IPython shell connected to a developer service (for exploration/R&D)."""
    with RemoteServer(service_provider, service, remove_ssl_context) as service:
        service.shell()


@cli.command()
@async_command
async def screenshot(service_provider: ServiceProviderDep, out: Path) -> None:
    """Capture a PNG screenshot (Depcrecated API)."""
    out.write_bytes(await ScreenshotService(lockdown=service_provider).take_screenshot())



import asyncio
import contextlib
import logging
import plistlib
import socket
import struct
from enum import Enum
from typing import Optional

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import (
    ConnectionFailedError,
    ConnectionTerminatedError,
    NoDeviceConnectedError,
    PyMobileDevice3Exception,
)
from pymobiledevice3.service_connection import ServiceConnection

CTRL_PORT = 0x43A  # 1082
CTRLCMD = b"BeginCtrl\0"
HELLOCMD = b"HelloConn\0"

FDR_SYNC_MSG = 0x1
FDR_PROXY_MSG = 0x105
FDR_PLIST_MSG = 0xBBAA
CHUNK_SIZE = 1048576

conn_port = None

logger = logging.getLogger(__name__)


class fdr_type(Enum):
    FDR_CTRL = 1
    FDR_CONN = 2


class FDRClient:
    SERVICE_PORT = CTRL_PORT
    ctrlprotoversion = 2

    def __init__(self, service: ServiceConnection) -> None:
        self.service = service
        self._conn_listener_task: Optional[asyncio.Task] = None

    @classmethod
    async def create(cls, type_: fdr_type, udid: Optional[str] = None) -> "FDRClient":
        device = await usbmux.select_device(udid)
        if device is None:
            if udid:
                raise ConnectionFailedError()
            raise NoDeviceConnectedError()

        logger.debug("connecting to FDR")
        if type_ == fdr_type.FDR_CTRL:
            service = await ServiceConnection.create_using_usbmux(
                device.serial, cls.SERVICE_PORT, connection_type="USB"
            )
        else:
            if conn_port is None:
                raise PyMobileDevice3Exception("FDR connection port is unavailable")
            service = await ServiceConnection.create_using_usbmux(device.serial, conn_port, connection_type="USB")

        client = cls(service)
        if type_ == fdr_type.FDR_CTRL:
            await client.ctrl_handshake()
        else:
            await client.sync_handshake()

        logger.debug("FDR connected")
        return client

    async def recv_plist(self) -> dict:
        return await self.service.recv_plist(endianity="<")

    async def send_recv_plist(self, plist: dict) -> dict:
        await self.service.send_plist(plist, endianity="<", fmt=plistlib.FMT_BINARY)
        return await self.service.recv_plist(endianity="<")

    async def ctrl_handshake(self) -> None:
        global conn_port

        logger.debug("About to do ctrl handshake")
        await self.service.sendall(CTRLCMD)

        if self.ctrlprotoversion != 2:
            raise NotImplementedError("TODO")

        req = {
            "Command": CTRLCMD,
            "CtrlProtoVersion": self.ctrlprotoversion,
        }
        resp = await self.send_recv_plist(req)
        conn_port = resp["ConnPort"]
        logger.debug(f"Ctrl handshake done (ConnPort = {conn_port})")

    async def sync_handshake(self) -> None:
        await self.service.sendall(HELLOCMD)

        if self.ctrlprotoversion != 2:
            raise NotImplementedError("TODO")

        reply = await self.recv_plist()
        cmd = reply["Command"]
        identifier = reply["Identifier"]

        if cmd != "HelloConn":
            raise PyMobileDevice3Exception("Did not receive HelloConn reply...")
        if identifier:
            logger.debug(f"got device identifier: {identifier}")

    async def handle_sync_cmd(self) -> None:
        await self.service.recvall(2)
        logger.debug("FDR connected in reply to sync message, starting command task")
        self._conn_listener_task = asyncio.create_task(run_fdr_listener(fdr_type.FDR_CONN), name="FDR-CONN")

    async def _proxy_service_to_host(self, host_socket: socket.socket) -> None:
        loop = asyncio.get_running_loop()
        while True:
            buf = await asyncio.to_thread(self.service.recv_sync, CHUNK_SIZE)
            if not buf:
                return
            await loop.sock_sendall(host_socket, buf)

    async def _proxy_host_to_service(self, host_socket: socket.socket) -> None:
        loop = asyncio.get_running_loop()
        while True:
            buf = await loop.sock_recv(host_socket, CHUNK_SIZE)
            if not buf:
                return
            await self.service.sendall(buf)

    async def handle_proxy_cmd(self) -> None:
        buf = await asyncio.to_thread(self.service.recv_sync, CHUNK_SIZE)
        logger.debug(f"got proxy command with {len(buf)} bytes")

        # acknowledge request and payload
        await self.service.sendall(struct.pack("<H", 5))
        if len(buf) < 3:
            logger.debug(f"FDR {self} proxy command data too short, retrying")
            return
        await self.service.sendall(buf)

        host = None
        port = None
        if buf[0] == 0 and buf[1] == 3:
            port = struct.unpack(">H", buf[-2:])[0]
            hostlen = buf[2]
            host = buf[3 : 3 + hostlen].decode()
            logger.debug(f"FDR {self} Proxy connect request to {host}:{port}")
        else:
            return

        host_socket = socket.socket()
        host_socket.setblocking(False)
        loop = asyncio.get_running_loop()
        await loop.sock_connect(host_socket, (host, port))

        service_to_host = asyncio.create_task(self._proxy_service_to_host(host_socket), name="FDR-proxy-s2h")
        host_to_service = asyncio.create_task(self._proxy_host_to_service(host_socket), name="FDR-proxy-h2s")

        done, pending = await asyncio.wait({service_to_host, host_to_service}, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        await asyncio.gather(*pending, return_exceptions=True)
        for task in done:
            with contextlib.suppress(Exception):
                task.result()

        host_socket.close()
        await self.service.close()

    async def handle_plist_cmd(self) -> None:
        d = await self.recv_plist()
        command = d["Command"]

        if command == "Ping":
            await self.send_recv_plist({"Pong": True})
        else:
            logger.warning(f"FDR {self} received unknown plist command: {command}")

    async def poll_and_handle_message(self) -> None:
        cmd = struct.unpack("<H", await self.service.recvall(2))[0]
        handlers = {
            FDR_SYNC_MSG: self.handle_sync_cmd,
            FDR_PROXY_MSG: self.handle_proxy_cmd,
            FDR_PLIST_MSG: self.handle_plist_cmd,
        }
        handler = handlers.get(cmd)
        if handler is None:
            logger.warning(f"ignoring FDR message: {cmd}")
            return
        await handler()


async def run_fdr_listener(type_: fdr_type, udid: Optional[str] = None) -> None:
    client: Optional[FDRClient] = None
    closing_via_generator_exit = False
    try:
        client = await FDRClient.create(type_, udid=udid)
        logger.debug(f"FDR {client} waiting for message...")
        while True:
            await client.poll_and_handle_message()
    except GeneratorExit:
        # Coroutine finalization cannot perform async cleanup (no await allowed).
        closing_via_generator_exit = True
        if client is not None and client.service.socket is not None:
            with contextlib.suppress(Exception):
                client.service.socket.close()
        raise
    except ConnectionTerminatedError:
        pass
    finally:
        if client is not None and not closing_via_generator_exit:
            await client.service.close()
    logger.debug(f"FDR {client} terminating...")


def start_fdr_task(type_: fdr_type, udid: Optional[str] = None) -> asyncio.Task:
    return asyncio.create_task(run_fdr_listener(type_, udid=udid), name=f"FDR-{type_.name}")



import logging

from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, print_json
from pymobiledevice3.services.os_trace import OsTraceService

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="processes",
    help="View process list using diagnosticsd API",
    no_args_is_help=True,
)


@cli.command("ps")
def processes_ps(service_provider: ServiceProviderDep) -> None:
    """show process list"""
    print_json(OsTraceService(lockdown=service_provider).get_pid_list().get("Payload"))


@cli.command("pgrep")
def processes_pgrep(service_provider: ServiceProviderDep, expression: str) -> None:
    """try to match processes pid by given expression (like pgrep)"""
    processes_list = OsTraceService(lockdown=service_provider).get_pid_list().get("Payload")
    for pid, process_info in processes_list.items():
        process_name = process_info.get("ProcessName")
        if expression in process_name:
            logger.info(f"{pid} {process_name}")



import logging
from pathlib import Path
from typing import Annotated

import typer
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command
from pymobiledevice3.services.simulate_location import DtSimulateLocation

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="simulate-location",
    help="Simulate GPS location (set, clear, or replay GPX routes).",
    no_args_is_help=True,
)


@cli.command("clear")
@async_command
async def simulate_location_clear(service_provider: ServiceProviderDep) -> None:
    """Stop location simulation and resume real GPS."""
    await DtSimulateLocation(service_provider).clear()


@cli.command("set")
@async_command
async def simulate_location_set(service_provider: ServiceProviderDep, latitude: float, longitude: float) -> None:
    """
    Set a fixed simulated location (latitude, longitude).
    Example: `set 40.690008 -74.045843` (Liberty Island).
    """
    await DtSimulateLocation(service_provider).set(latitude, longitude)


@cli.command("play")
@async_command
async def simulate_location_play(
    service_provider: ServiceProviderDep,
    filename: Annotated[
        Path,
        typer.Argument(exists=True, file_okay=True, dir_okay=False),
    ],
    timing_randomness_range: int,
    disable_sleep: Annotated[bool, typer.Option()] = False,
) -> None:
    """Replay a GPX route; optionally disable sleeps and add timing jitter."""
    await DtSimulateLocation(service_provider).play_gpx_file(
        str(filename),
        disable_sleep=disable_sleep,
        timing_randomness_range=timing_randomness_range,
    )



import asyncio
import hashlib
import logging
import os
import plistlib
import typing

from tqdm import trange

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.service_connection import ServiceConnection

ASR_VERSION = 1
ASR_STREAM_ID = 1
DEFAULT_ASR_SYNC_PORT = 12345
ASR_FEC_SLICE_STRIDE = 40
ASR_PACKETS_PER_FEC = 25
ASR_PAYLOAD_PACKET_SIZE = 1450
ASR_PAYLOAD_CHUNK_SIZE = 0x20000
ASR_CHECKSUM_CHUNK_SIZE = ASR_PAYLOAD_CHUNK_SIZE

logger = logging.getLogger(__name__)


class ASRClient:
    """
    ASR — Apple Software Restore
    """

    def __init__(self, udid: str) -> None:
        self._udid: str = udid
        self.logger = logging.getLogger(f"{asyncio.current_task().get_name()}-{__name__}")
        self.service: typing.Optional[ServiceConnection] = None
        self.checksum_chunks: bool = False

    async def connect(self, port: int = DEFAULT_ASR_SYNC_PORT) -> None:
        self.service = await ServiceConnection.create_using_usbmux(self._udid, port, connection_type="USB")
        await self.service.start()

        # receive Initiate command message
        data = await self.recv_plist()
        self.logger.debug(f"got command: {data}")

        command = data.get("Command")
        if command != "Initiate":
            raise PyMobileDevice3Exception(f"invalid command received: {command}")

        self.checksum_chunks = data.get("Checksum Chunks", False)
        self.logger.debug(f"Checksum Chunks: {self.checksum_chunks}")

    async def recv_plist(self) -> dict:
        buf = b""
        while not buf.endswith(b"</plist>\n"):
            buf += await self.service.recvall(1)
        return plistlib.loads(buf)

    async def send_plist(self, plist: dict) -> None:
        self.logger.debug(plistlib.dumps(plist).decode())
        await self.send_buffer(plistlib.dumps(plist))

    async def send_buffer(self, buf: bytes) -> None:
        await self.service.sendall(buf)

    async def handle_oob_data_request(self, packet: dict, filesystem: typing.IO):
        oob_length = packet["OOB Length"]
        oob_offset = packet["OOB Offset"]
        filesystem.seek(oob_offset, os.SEEK_SET)

        oob_data = filesystem.read(oob_length)
        assert len(oob_data) == oob_length

        await self.send_buffer(oob_data)

    async def perform_validation(self, filesystem: typing.IO) -> None:
        filesystem.seek(0, os.SEEK_END)
        length = filesystem.tell()
        filesystem.seek(0, os.SEEK_SET)

        payload_info = {
            "Port": 1,
            "Size": length,
        }

        packet_info = {}
        if self.checksum_chunks:
            packet_info["Checksum Chunk Size"] = ASR_CHECKSUM_CHUNK_SIZE

        packet_info["FEC Slice Stride"] = ASR_FEC_SLICE_STRIDE
        packet_info["Packet Payload Size"] = ASR_PAYLOAD_PACKET_SIZE
        packet_info["Packets Per FEC"] = ASR_PACKETS_PER_FEC
        packet_info["Payload"] = payload_info
        packet_info["Stream ID"] = ASR_STREAM_ID
        packet_info["Version"] = ASR_VERSION

        await self.send_plist(packet_info)

        while True:
            packet = await self.recv_plist()
            self.logger.debug(f"perform_validation: {packet}")
            command = packet["Command"]

            if command == "Payload":
                break

            elif command == "OOBData":
                await self.handle_oob_data_request(packet, filesystem)

            else:
                raise PyMobileDevice3Exception(f"unknown packet: {packet}")

    async def send_payload(self, filesystem: typing.IO) -> None:
        filesystem.seek(0, os.SEEK_END)
        length = filesystem.tell()
        filesystem.seek(0, os.SEEK_SET)

        for _ in trange(0, length, ASR_PAYLOAD_CHUNK_SIZE, dynamic_ncols=True):
            chunk = filesystem.read(ASR_PAYLOAD_CHUNK_SIZE)

            if self.checksum_chunks:
                chunk += hashlib.sha1(chunk).digest()

            await self.send_buffer(chunk)

    async def close(self) -> None:
        await self.service.close()



import asyncio

import pytest

from pymobiledevice3.exceptions import DvtDirListError, UnrecognizedSelectorError
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.application_listing import ApplicationListing
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl


async def get_process_data(dvt: DvtSecureSocketProxyService, name: str):
    processes = await DeviceInfo(dvt).proclist()
    return next(process for process in processes if process["name"] == name)


@pytest.mark.asyncio
async def test_ls(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test listing a directory.
    """
    ls = set(await DeviceInfo(dvt).ls("/"))
    assert {"usr", "bin", "etc", "var", "private", "Applications", "Developer"} <= ls


@pytest.mark.asyncio
async def test_ls_failure(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test listing a directory.
    """
    with pytest.raises(DvtDirListError):
        await DeviceInfo(dvt).ls("Directory that does not exist")


@pytest.mark.asyncio
async def test_proclist(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test listing processes.
    """
    lockdownd = await get_process_data(dvt, "lockdownd")
    assert lockdownd["realAppName"] == "/usr/libexec/lockdownd"
    assert not lockdownd["isApplication"]


@pytest.mark.asyncio
async def test_applist(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test listing applications.
    """
    apps = await ApplicationListing(dvt).applist()
    safari = next(app for app in apps if app["DisplayName"] == "StocksWidget")
    assert safari["CFBundleIdentifier"] == "com.apple.stocks.widget"
    assert safari["Restricted"] == 1
    assert safari["Type"] == "PluginKit"


@pytest.mark.asyncio
async def test_memlimitoff(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test disabling memory limit.
    """
    process = await get_process_data(dvt, "SpringBoard")
    await ProcessControl(dvt).disable_memory_limit_for_pid(process["pid"])


@pytest.mark.asyncio
async def test_kill(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test killing a process.
    """
    aggregated = await get_process_data(dvt, "SpringBoard")
    await ProcessControl(dvt).kill(aggregated["pid"])
    # give the os some time to start the process again
    await asyncio.sleep(3)
    aggregated_after_kill = await get_process_data(dvt, "SpringBoard")
    if "startDate" in aggregated:
        assert aggregated["startDate"] < aggregated_after_kill["startDate"]


@pytest.mark.asyncio
async def test_launch(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test launching a process.
    """
    pid = await ProcessControl(dvt).launch("com.apple.mobilesafari")
    assert pid
    for process in await DeviceInfo(dvt).proclist():
        if pid == process["pid"]:
            assert process["name"] == "MobileSafari"


@pytest.mark.asyncio
async def test_system_information(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test getting system information.
    """
    try:
        system_info = await DeviceInfo(dvt).system_information()
    except UnrecognizedSelectorError:
        pytest.skip("device doesn't support this method")
    assert "_deviceDescription" in system_info and system_info["_deviceDescription"].startswith("Build Version")


@pytest.mark.asyncio
async def test_hardware_information(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test getting hardware information.
    """
    hardware_info = await DeviceInfo(dvt).hardware_information()
    assert hardware_info["numberOfCpus"] > 0


@pytest.mark.asyncio
async def test_network_information(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test getting network information.
    """
    network_info = await DeviceInfo(dvt).network_information()
    assert network_info["lo0"] == "Loopback"



import pytest

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.web_protocol.driver import WebDriver
from pymobiledevice3.services.webinspector import SAFARI, WebinspectorService


@pytest.fixture
async def webdriver(lockdown: LockdownClient):
    inspector = WebinspectorService(lockdown=lockdown)
    await inspector.connect()
    safari = await inspector.open_app(SAFARI)
    await inspector.flush_input(1)
    session = await inspector.automation_session(safari)
    driver = WebDriver(session)
    await driver.start_session()
    try:
        yield driver
    finally:
        await inspector.close()



import logging

from ipsw_parser.build_identity import BuildIdentity
from pyimg4 import IM4P, IM4R, IMG4, RestoreProperty

logger = logging.getLogger(__name__)

COMPONENT_FOURCC = {
    "ACIBT": "acib",
    "ACIBTLPEM": "lpbt",
    "ACIWIFI": "aciw",
    "ANE": "anef",
    "ANS": "ansf",
    "AOP": "aopf",
    "AVE": "avef",
    "Alamo": "almo",
    "Ap,ANE1": "ane1",
    "Ap,ANE2": "ane2",
    "Ap,ANE3": "ane3",
    "Ap,AudioAccessibilityBootChime": "auac",
    "Ap,AudioBootChime": "aubt",
    "Ap,AudioPowerAttachChime": "aupr",
    "Ap,BootabilityBrainTrustCache": "trbb",
    "Ap,CIO": "ciof",
    "Ap,HapticAssets": "hpas",
    "Ap,LocalBoot": "lobo",
    "Ap,LocalPolicy": "lpol",
    "Ap,NextStageIM4MHash": "nsih",
    "Ap,RecoveryOSPolicyNonceHash": "ronh",
    "Ap,RestoreANE1": "ran1",
    "Ap,RestoreANE2": "ran2",
    "Ap,RestoreANE3": "ran3",
    "Ap,RestoreCIO": "rcio",
    "Ap,RestoreDCP2": "rdc2",
    "Ap,RestoreTMU": "rtmu",
    "Ap,Scorpius": "scpf",
    "Ap,SystemVolumeCanonicalMetadata": "msys",
    "Ap,TMU": "tmuf",
    "Ap,VolumeUUID": "vuid",
    "Ap,rOSLogo1": "rlg1",
    "Ap,rOSLogo2": "rlg2",
    "AppleLogo": "logo",
    "AudioCodecFirmware": "acfw",
    "BatteryCharging": "glyC",
    "BatteryCharging0": "chg0",
    "BatteryCharging1": "chg1",
    "BatteryFull": "batF",
    "BatteryLow0": "bat0",
    "BatteryLow1": "bat1",
    "BatteryPlugin": "glyP",
    "CFELoader": "cfel",
    "CrownFirmware": "crwn",
    "DCP": "dcpf",
    "Dali": "dali",
    "DeviceTree": "dtre",
    "Diags": "diag",
    "EngineeringTrustCache": "dtrs",
    "ExtDCP": "edcp",
    "GFX": "gfxf",
    "Hamm": "hamf",
    "Homer": "homr",
    "ISP": "ispf",
    "InputDevice": "ipdf",
    "KernelCache": "krnl",
    "LLB": "illb",
    "LeapHaptics": "lphp",
    "Liquid": "liqd",
    "LoadableTrustCache": "ltrs",
    "LowPowerWallet0": "lpw0",
    "LowPowerWallet1": "lpw1",
    "LowPowerWallet2": "lpw2",
    "MacEFI": "mefi",
    "MtpFirmware": "mtpf",
    "Multitouch": "mtfw",
    "NeedService": "nsrv",
    "OS": "OS\0\0",
    "OSRamdisk": "osrd",
    "PEHammer": "hmmr",
    "PERTOS": "pert",
    "PHLEET": "phlt",
    "PMP": "pmpf",
    "PersonalizedDMG": "pdmg",
    "RBM": "rmbt",
    "RTP": "rtpf",
    "Rap,SoftwareBinaryDsp1": "sbd1",
    "Rap,RTKitOS": "rkos",
    "Rap,RestoreRTKitOS": "rrko",
    "RecoveryMode": "recm",
    "RestoreANS": "rans",
    "RestoreDCP": "rdcp",
    "RestoreDeviceTree": "rdtr",
    "RestoreExtDCP": "recp",
    "RestoreKernelCache": "rkrn",
    "RestoreLogo": "rlgo",
    "RestoreRTP": "rrtp",
    "RestoreRamDisk": "rdsk",
    "RestoreSEP": "rsep",
    "RestoreTrustCache": "rtsc",
    "SCE": "scef",
    "SCE1Firmware": "sc1f",
    "SEP": "sepi",
    "SIO": "siof",
    "StaticTrustCache": "trst",
    "SystemLocker": "lckr",
    "SystemVolume": "isys",
    "WCHFirmwareUpdater": "wchf",
    "ftap": "ftap",
    "ftsp": "ftsp",
    "iBEC": "ibec",
    "iBSS": "ibss",
    "iBoot": "ibot",
    "iBootData": "ibdt",
    "iBootDataStage1": "ibd1",
    "iBootTest": "itst",
    "rfta": "rfta",
    "rfts": "rfts",
    "Ap,DCP2": "dcp2",
    "Ap,RestoreSecureM3Firmware": "rsm3",
    "Ap,RestoreSecurePageTableMonitor": "rspt",
    "Ap,RestoreTrustedExecutionMonitor": "rtrx",
    "Ap,RestorecL4": "rxcl",
}


def stitch_component(
    name: str, im4p_data: bytes, tss: dict, build_identity: BuildIdentity, ap_parameters: dict
) -> bytes:
    logger.info(f"Personalizing IMG4 component {name}...")

    im4p = IM4P(data=im4p_data)

    # patch fourcc
    fourcc = COMPONENT_FOURCC.get(name)
    if fourcc is not None:
        im4p.fourcc = fourcc

    # check if we have a *-TBM entry for the given component
    tbm_dict = tss.get(f"{name}-TBM")

    info = build_identity["Info"]
    nonce_slot_required = info.get("RequiresNonceSlot", False) and name in ("SEP", "SepStage1", "LLB")

    im4r = None
    if tbm_dict is not None or nonce_slot_required:
        im4r = IM4R()

        if nonce_slot_required:
            logger.debug(f"{name}: RequiresNonceSlot for {name}")
            if name in ("SEP", "SepStage1"):
                snid = ap_parameters.get("SepNonceSlotID", info.get("SepNonceSlotID", 2))
                logger.debug(f"snid: {snid}")
                im4r.add_property(RestoreProperty(fourcc="snid", value=snid))
            else:
                anid = ap_parameters.get("ApNonceSlotID", info.get("ApNonceSlotID", 0))
                logger.debug(f"anid: {anid}")
                im4r.add_property(RestoreProperty(fourcc="anid", value=anid))

    if tbm_dict is not None:
        for key in tbm_dict:
            logger.debug(f"{name}: Adding property {key}")
            im4r.add_property(RestoreProperty(fourcc=key, value=tbm_dict[key]))

    return IMG4(im4p=im4p, im4m=tss.ap_img4_ticket, im4r=im4r).output()



import asyncio
import contextlib
from base64 import b64decode, b64encode
from datetime import datetime
from io import BytesIO
from typing import Optional

from PIL import Image


class ScreenCast:
    def __init__(self, target, format_: str, quality: int, max_width: int, max_height: int):
        """
        :param pymobiledevice3.services.web_protocol.cdp_target.CdpTarget target:
        :param format_: Image compression format. Allowed values: jpeg, png.
        :param quality: Compression quality from range [0..100].
        :param max_width: Maximum screenshot width.
        :param max_height: Maximum screenshot height.
        """
        self.target = target
        self.format_ = format_
        self.quality = quality
        self.max_width = max_width
        self.max_height = max_height
        self.frames_acked = []
        self.frame_id = 1
        self.frame_interval = 250
        self.device_width = 0
        self.device_height = 0
        self.page_scale_factor = 0
        self._run = True
        self.recording_task: Optional[asyncio.Task] = None

    def get_scale(self) -> float:
        """The amount screen pixels in one devtools pixel."""
        real_height = self.device_height * self.page_scale_factor
        real_width = self.device_width * self.page_scale_factor
        return min(self.max_height / real_height, self.max_width / real_width, 1) * self.page_scale_factor

    def get_scaled_width(self) -> int:
        """Width of screenshot after scaling."""
        return int(self.get_scale() * self.device_width)

    def get_scaled_height(self) -> int:
        """Height of screenshot after scaling."""
        return int(self.get_scale() * self.device_height)

    async def start(self, message_id: int):
        """
        Start sending screenshots to the devtools.
        :param message_id: Message id to use when requesting WIR data concerning the screencast.
        """
        device_size = await self.target.evaluate_and_result(
            message_id,
            (
                '(window.innerWidth > 0 ? window.innerWidth : screen.width) + "," + '
                '(window.innerHeight > 0 ? window.innerHeight : screen.height) + "," + '
                "window.devicePixelRatio"
            ),
        )
        self.device_width, self.device_height, self.page_scale_factor = list(map(int, device_size.split(",")))
        self._run = True
        self.recording_task = asyncio.create_task(self.recording_loop(message_id))

    async def stop(self):
        """Stop sending screenshots to the devtools."""
        self._run = False
        self.recording_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self.recording_task
        self.recording_task = None

    def ack(self, frame_id: int):
        """Handle acknowledgement for screencast frames."""
        self.frames_acked.append(frame_id)

    def resize_jpeg(self, data: str) -> str:
        """
        Resize a screenshot to fit the devtools requested size.
        :param data: Base64 of JPEG data.
        :return: Base 64 of resized JPEG data.
        """
        resized_img = Image.open(BytesIO(b64decode(data)))
        resized_img = resized_img.resize((self.get_scaled_width(), self.get_scaled_height()), Image.ANTIALIAS)
        resized_img = resized_img.convert("RGB")
        resized = BytesIO()
        resized_img.save(resized, format="jpeg", quality="maximum")
        return b64encode(resized.getvalue()).decode()

    async def get_offsets(self, message_id: int):
        """
        Get the offset of the screenshot from the start of the page.
        :param message_id: Message id to use when requesting WIR data concerning the screencast.
        :return: Tuple of (offsetTop, pageXOffset, pageYOffset).
        :rtype: tuple
        """
        frame_size = await self.target.evaluate_and_result(
            message_id, 'window.document.body.offsetTop + "," + window.pageXOffset + "," + window.pageYOffset'
        )
        if frame_size is None or not isinstance(frame_size, str):
            return 0, 0, 0
        return tuple(map(int, frame_size.split(",")))

    async def recording_loop(self, message_id):
        """
        Fetch screenshots and send to devtools.
        :param message_id: Message id to use when requesting WIR data concerning the screencast.
        """
        while self._run:
            await asyncio.sleep(self.frame_interval / 1000)
            if self.frame_id > 1 and (self.frame_id - 1) not in self.frames_acked:
                continue
            self.frame_id += 1
            offset_top, scroll_offset_x, scroll_offset_y = await self.get_offsets(message_id)
            event = await self.target.send_message_with_result(
                message_id,
                "Page.snapshotRect",
                {
                    "x": 0,
                    "y": 0,
                    "width": self.device_width,
                    "height": self.device_height,
                    "coordinateSystem": "Viewport",
                },
            )
            data = event["result"]["dataURL"]
            data = data[data.find("base64,") + 7 :]
            await self.target.output_queue.put({
                "method": "Page.screencastFrame",
                "params": {
                    "data": self.resize_jpeg(data),
                    "sessionId": self.frame_id - 1,
                    "metadata": {
                        "pageScaleFactor": self.page_scale_factor,
                        "offsetTop": offset_top,
                        "deviceWidth": self.get_scaled_width(),
                        "deviceHeight": self.get_scaled_height(),
                        "scrollOffsetX": scroll_offset_x,
                        "scrollOffsetY": scroll_offset_y,
                        "timestamp": datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f"),
                    },
                },
            })



from pathlib import Path

import pytest
from packaging.version import Version

from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.remote.remote_service_discovery import RemoteServiceDiscoveryService
from pymobiledevice3.services.dtfetchsymbols import DtFetchSymbols
from pymobiledevice3.services.remote_fetch_symbols import RemoteFetchSymbolsService


@pytest.mark.asyncio
async def test_fetch_symbols_list(service_provider: LockdownServiceProvider) -> None:
    """
    Test listing of device symbol files
    """
    if Version(service_provider.product_version) < Version("17.0"):
        assert len(await DtFetchSymbols(service_provider).list_files()) > 0
    else:
        if not isinstance(service_provider, RemoteServiceDiscoveryService):
            pytest.skip("requires RemoteServiceDiscoveryService")

        async with RemoteFetchSymbolsService(service_provider) as fetch_symbols:
            assert len(await fetch_symbols.get_dsc_file_list()) > 0


@pytest.mark.asyncio
async def test_fetch_symbols_download(service_provider: LockdownServiceProvider, tmp_path: Path) -> None:
    """
    Test that the download mechanism can transfer data from the device.
    Only probes a small amount to verify the transfer protocol works without
    downloading entire symbol files (which can be several GB).
    """
    _PROBE_BYTES = 64 * 1024  # 64 KB is enough to confirm data flows

    if Version(service_provider.product_version) < Version("17.0"):
        tmp_file = Path(tmp_path) / "tmp"
        with tmp_file.open("wb") as file:
            await DtFetchSymbols(service_provider).get_file(0, file, max_bytes=_PROBE_BYTES)
        assert tmp_file.stat().st_size > 0
    else:
        if not isinstance(service_provider, RemoteServiceDiscoveryService):
            pytest.skip("requires RemoteServiceDiscoveryService")

        async with RemoteFetchSymbolsService(service_provider) as fetch_symbols:
            files = await fetch_symbols.get_dsc_file_list()
            assert len(files) > 0
            # Receive just the first chunk of the first file to confirm data transfer works.
            received = 0
            async for chunk in fetch_symbols.service.iter_file_chunks(files[0].file_size, file_idx=0):
                received += len(chunk)
                break
            assert received > 0



import ipaddress
import logging
from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import Union

from construct import Adapter, Bytes, Int8ul, Int16ub, Int32ul, Switch, this
from construct_typed import DataclassMixin, TStruct, csfield

from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments import ChannelService


class IpAddressAdapter(Adapter):
    """Decode raw address bytes into ipaddress objects."""

    def _decode(self, obj, context, path):
        return ipaddress.ip_address(obj)


@dataclass
class AddressV6(DataclassMixin):
    """IPv6 payload for a socket address."""

    flow_info: int = csfield(Int32ul)
    address: ipaddress.IPv6Address = csfield(IpAddressAdapter(Bytes(16)))
    scope_id: int = csfield(Int32ul)


@dataclass
class AddressV4(DataclassMixin):
    """IPv4 payload for a socket address."""

    address: ipaddress.IPv4Address = csfield(IpAddressAdapter(Bytes(4)))
    _zero: bytes = csfield(Bytes(8))


@dataclass
class SocketAddress(DataclassMixin):
    """Parsed socket address with family-specific payload."""

    length: int = csfield(Int8ul)
    family: int = csfield(Int8ul)
    port: int = csfield(Int16ub)
    data: Union[AddressV4, AddressV6] = csfield(
        Switch(
            this.length,
            {
                0x1C: TStruct(AddressV6),
                0x10: TStruct(AddressV4),
            },
        )
    )


address_t = TStruct(SocketAddress)

MESSAGE_TYPE_INTERFACE_DETECTION = 0
MESSAGE_TYPE_CONNECTION_DETECTION = 1
MESSAGE_TYPE_CONNECTION_UPDATE = 2


@dataclass
class InterfaceDetectionEvent:
    """Interface detection event emitted by Instruments."""

    interface_index: int
    name: str


@dataclass
class ConnectionDetectionEvent:
    """Connection detection event emitted by Instruments."""

    local_address: SocketAddress
    remote_address: SocketAddress
    interface_index: int
    pid: int
    recv_buffer_size: int
    recv_buffer_used: int
    serial_number: int
    kind: int


@dataclass
class ConnectionUpdateEvent:
    """Connection update event emitted by Instruments."""

    rx_packets: int
    rx_bytes: int
    tx_packets: int
    tx_bytes: int
    rx_dups: int
    rx000: int
    tx_retx: int
    min_rtt: int
    avg_rtt: int
    connection_serial: int
    time: int


NetworkMonitorEvent = Union[InterfaceDetectionEvent, ConnectionDetectionEvent, ConnectionUpdateEvent]


class NetworkMonitor(ChannelService):
    """Iterate over network monitoring events from the Instruments service."""

    IDENTIFIER = "com.apple.instruments.server.services.networking"

    def __init__(self, dvt: DvtSecureSocketProxyService):
        self.logger = logging.getLogger(__name__)
        super().__init__(dvt)

    async def __aenter__(self) -> "NetworkMonitor":
        channel = await self._channel_ref()
        await channel.startMonitoring(expects_reply=False)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        channel = await self._channel_ref()
        await channel.stopMonitoring()

    async def __aiter__(self) -> AsyncIterator[NetworkMonitorEvent]:
        """Yield network events as they arrive from the service."""

        while True:
            message = await self._receive_message()

            event = None

            if message is None:
                continue

            if message[0] == MESSAGE_TYPE_INTERFACE_DETECTION:
                event = InterfaceDetectionEvent(*message[1])
            elif message[0] == MESSAGE_TYPE_CONNECTION_DETECTION:
                (
                    local_address,
                    remote_address,
                    interface_index,
                    pid,
                    recv_buffer_size,
                    recv_buffer_used,
                    serial_number,
                    kind,
                ) = message[1]
                event = ConnectionDetectionEvent(
                    local_address=address_t.parse(local_address),
                    remote_address=address_t.parse(remote_address),
                    interface_index=interface_index,
                    pid=pid,
                    recv_buffer_size=recv_buffer_size,
                    recv_buffer_used=recv_buffer_used,
                    serial_number=serial_number,
                    kind=kind,
                )
            elif message[0] == MESSAGE_TYPE_CONNECTION_UPDATE:
                event = ConnectionUpdateEvent(*message[1])
            else:
                self.logger.warning(f"unsupported event type: {message[0]}")
            yield event

    async def _receive_message(self):
        channel = await self._channel_ref()
        return await channel.receive_plist()



from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.location_simulation import LocationSimulation


async def test_set_location(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test set location.
    """
    # set to liberty island
    await LocationSimulation(dvt).set(40.690008, -74.045843)


async def test_clear_location(dvt: DvtSecureSocketProxyService) -> None:
    """
    Test clear location simulation
    """
    # set to liberty island
    await LocationSimulation(dvt).clear()



from collections.abc import AsyncGenerator

from pymobiledevice3.exceptions import ConnectionTerminatedError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.lockdown_service import LockdownService
from pymobiledevice3.utils import try_decode

CHUNK_SIZE = 4096
TIME_FORMAT = "%H:%M:%S"
SYSLOG_LINE_SPLITTER = b"\n\x00"


class SyslogService(LockdownService):
    """
    View system logs
    """

    SERVICE_NAME = "com.apple.syslog_relay"
    RSD_SERVICE_NAME = "com.apple.syslog_relay.shim.remote"

    def __init__(self, service_provider: LockdownServiceProvider):
        if isinstance(service_provider, LockdownClient):
            super().__init__(service_provider, self.SERVICE_NAME)
        else:
            super().__init__(service_provider, self.RSD_SERVICE_NAME)

    async def watch(self) -> AsyncGenerator[str, None]:
        buf = b""
        while True:
            # read in chunks till we have at least one syslog line
            chunk = await self.service.recv_any(CHUNK_SIZE)

            if len(chunk) == 0:
                raise ConnectionTerminatedError()

            buf += chunk

            # SYSLOG_LINE_SPLITTER is used to split each syslog line
            if SYSLOG_LINE_SPLITTER in buf:
                lines = buf.split(SYSLOG_LINE_SPLITTER)

                # handle partial last lines
                if not buf.endswith(SYSLOG_LINE_SPLITTER):
                    buf = lines[-1]
                    lines = lines[:-1]

                for line in lines:
                    if len(line) == 0:
                        continue

                    yield try_decode(line)



import asyncio
import contextlib
import logging
import os
import posixpath
import shlex
from datetime import datetime
from enum import IntEnum
from pathlib import Path
from typing import Annotated, NamedTuple, Optional

import typer
from click.exceptions import BadParameter, MissingParameter, UsageError
from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import ServiceProviderDep, async_command, print_json, user_requested_colored_output
from pymobiledevice3.cli.developer.dvt import core_profile_session, simulate_location, sysmon
from pymobiledevice3.exceptions import DvtDirListError, UnrecognizedSelectorError
from pymobiledevice3.services.dvt.dvt_secure_socket_proxy import DvtSecureSocketProxyService
from pymobiledevice3.services.dvt.instruments.activity_trace_tap import ActivityTraceTap, decode_message_format
from pymobiledevice3.services.dvt.instruments.application_listing import ApplicationListing
from pymobiledevice3.services.dvt.instruments.device_info import DeviceInfo
from pymobiledevice3.services.dvt.instruments.energy_monitor import EnergyMonitor
from pymobiledevice3.services.dvt.instruments.graphics import Graphics
from pymobiledevice3.services.dvt.instruments.network_monitor import ConnectionDetectionEvent, NetworkMonitor
from pymobiledevice3.services.dvt.instruments.notifications import Notifications
from pymobiledevice3.services.dvt.instruments.process_control import ProcessControl
from pymobiledevice3.services.dvt.instruments.screenshot import Screenshot
from pymobiledevice3.services.dvt.testmanaged.xcuitest import XCUITestListener, XCUITestService
from pymobiledevice3.services.remote_server import MessageAux
from pymobiledevice3.utils import run_in_loop

logger = logging.getLogger(__name__)


class MatchedProcessByPid(NamedTuple):
    name: str
    pid: int


cli = InjectingTyper(
    name="dvt",
    help="Drive DVT instrumentation APIs (process control, metrics, traces).",
    no_args_is_help=True,
)

cli.add_typer(sysmon.cli)
cli.add_typer(core_profile_session.cli)
cli.add_typer(simulate_location.cli)


@cli.command("proclist")
@async_command
async def proclist(service_provider: ServiceProviderDep) -> None:
    """Show processes (with start times) via DVT."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        processes = await DeviceInfo(dvt).proclist()
        for process in processes:
            if "startDate" in process:
                process["startDate"] = str(process["startDate"])

        print_json(processes)


@cli.command("is-running-pid")
@async_command
async def is_running_pid(service_provider: ServiceProviderDep, pid: int) -> None:
    """Check if a PID is currently running."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        print_json(await DeviceInfo(dvt).is_running_pid(pid))


@cli.command("memlimitoff")
@async_command
async def memlimitoff(service_provider: ServiceProviderDep, pid: int) -> None:
    """Disable jetsam memory limit for a PID."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        await ProcessControl(dvt).disable_memory_limit_for_pid(pid)


@cli.command("applist")
@async_command
async def applist(service_provider: ServiceProviderDep) -> None:
    """List installed applications via DVT."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        apps = await ApplicationListing(dvt).applist()
        print_json(apps)


class Signals(IntEnum):
    """Platform-independent version of `signal.Signals`, allowing names to be used on Windows."""

    HUP = 1
    INT = 2
    QUIT = 3
    ILL = 4
    TRAP = 5
    ABRT = 6
    EMT = 7
    FPE = 8
    KILL = 9
    BUS = 10
    SEGV = 11
    SYS = 12
    PIPE = 13
    ALRM = 14
    TERM = 15
    URG = 16
    STOP = 17
    TSTP = 18
    CONT = 19
    CHLD = 20
    TTIN = 21
    TTOU = 22
    IO = 23
    XCPU = 24
    XFSZ = 25
    VTALRM = 26
    PROF = 27
    WINCH = 28
    INFO = 29
    USR1 = 30
    USR2 = 31


@cli.command("signal")
@async_command
async def send_signal(
    service_provider: ServiceProviderDep,
    pid: int,
    sig: Annotated[
        Optional[int],
        typer.Argument(),
    ] = None,
    signal_name: Annotated[
        Optional[str],
        typer.Option("--signal-name", "-s"),
    ] = None,
) -> None:
    """Send a signal to a PID (choose numeric SIG or --signal-name)."""
    if sig is not None and signal_name is not None:
        raise UsageError(message="Cannot give SIG and SIGNAL-NAME together")

    if signal_name is not None:
        normalized_signal_name = signal_name.upper().removeprefix("SIG")
        try:
            sig = Signals[normalized_signal_name]
        except KeyError:
            raise BadParameter(f"{signal_name!r} is not a valid signal") from None
    elif sig is not None:
        try:
            sig = Signals(sig)
        except ValueError:
            raise BadParameter(f"{sig} is not a valid signal") from None
    else:
        raise MissingParameter(param_type="argument|option", param_hint="'SIG|SIGNAL-NAME'")

    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        await ProcessControl(dvt).signal(pid, sig)


@cli.command("kill")
@async_command
async def kill(service_provider: ServiceProviderDep, pid: int) -> None:
    """Kill a process by PID."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        await ProcessControl(dvt).kill(pid)


@cli.command()
@async_command
async def process_id_for_bundle_id(service_provider: ServiceProviderDep, app_bundle_identifier: str) -> None:
    """Get PID of a bundle identifier (only returns a valid value if its running)."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        print(await ProcessControl(dvt).process_identifier_for_bundle_identifier(app_bundle_identifier))


async def get_matching_processes(
    device_info: DeviceInfo,
    name: Optional[str] = None,
    bundle_identifier: Optional[str] = None,
) -> list[MatchedProcessByPid]:
    result: list[MatchedProcessByPid] = []
    for process in await device_info.proclist():
        current_name = process["name"]
        current_bundle_identifier = process.get("bundleIdentifier", "")
        pid = process["pid"]
        if (bundle_identifier is not None and bundle_identifier in current_bundle_identifier) or (
            name is not None and name in current_name
        ):
            result.append(MatchedProcessByPid(name=current_name, pid=pid))
    return result


@cli.command("pkill")
@async_command
async def pkill(
    service_provider: ServiceProviderDep,
    expressions: Annotated[
        list[str],
        typer.Argument(help="One or more process-name (or bundle id) expressions to match."),
    ],
    bundle: Annotated[
        bool,
        typer.Option(help="Treat given expressions as bundle-identifiers instead of process names"),
    ] = False,
) -> None:
    """Kill all processes containing each expression in their name."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        device_info = DeviceInfo(dvt)
        process_control = ProcessControl(dvt)

        for expression in expressions:
            matching_name = expression if not bundle else None
            matching_bundle_identifier = expression if bundle else None
            matching_processes = await get_matching_processes(
                device_info, name=matching_name, bundle_identifier=matching_bundle_identifier
            )

            for process in matching_processes:
                logger.info(f"Killing {process.name}({process.pid})")
                await process_control.kill(process.pid)


@cli.command("launch")
@async_command
async def launch(
    service_provider: ServiceProviderDep,
    arguments: str,
    kill_existing: Annotated[
        bool,
        typer.Option(help="Whether to kill an existing instance of this process"),
    ] = True,
    suspended: Annotated[
        bool,
        typer.Option(help="Same as WaitForDebugger"),
    ] = False,
    env: Annotated[
        Optional[list[str]],
        typer.Option(
            help="Environment variable to pass to process given as key=value (can be specified multiple times)"
        ),
    ] = None,
    stream: bool = False,
) -> None:
    """Launch a process."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        parsed_arguments = shlex.split(arguments)
        process_control = ProcessControl(dvt)
        pid = await process_control.launch(
            bundle_id=parsed_arguments[0],
            arguments=parsed_arguments[1:],
            kill_existing=kill_existing,
            start_suspended=suspended,
            environment=dict(var.split("=", 1) for var in env or ()),
        )
        print(f"Process launched with pid {pid}")
        if stream:
            async for output_received in process_control:
                logging.getLogger(f"PID:{output_received.pid}").info(output_received.message.strip())


@cli.command("shell")
def dvt_shell(service_provider: ServiceProviderDep) -> None:
    """Launch developer shell (used for pymobiledevice3 R&D)"""
    dvt = DvtSecureSocketProxyService(lockdown=service_provider)
    run_in_loop(dvt.perform_handshake())
    dvt.shell()


async def show_dirlist(channel, dirname: str, recursive: bool = False) -> None:
    try:
        await channel.directoryListingForPath_(MessageAux().append_obj(dirname))
        filenames = await channel.receive_plist()
        if filenames is None:
            raise DvtDirListError()
    except DvtDirListError:
        return

    for filename in filenames:
        filename = posixpath.join(dirname, filename)
        print(filename)
        if recursive:
            await show_dirlist(channel, filename, recursive=recursive)


@cli.command("ls")
@async_command
async def ls(
    service_provider: ServiceProviderDep,
    path: str,
    recursive: Annotated[
        bool,
        typer.Option("--recursive", "-r"),
    ] = False,
) -> None:
    """List directory"""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        channel = await dvt.make_channel(DeviceInfo.IDENTIFIER)
        await show_dirlist(channel, path, recursive=recursive)


@cli.command("device-information")
@async_command
async def device_information(service_provider: ServiceProviderDep) -> None:
    """Print system information"""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        device_info = DeviceInfo(dvt)
        info = {
            "hardware": await device_info.hardware_information(),
            "network": await device_info.network_information(),
            "kernel-name": await device_info.mach_kernel_name(),
            "kpep-database": await device_info.kpep_database(),
        }
        with contextlib.suppress(UnrecognizedSelectorError):
            info["system"] = await device_info.system_information()
        print_json(info)


@cli.command("netstat")
@async_command
async def netstat(service_provider: ServiceProviderDep) -> None:
    """Print information about current network activity."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt, NetworkMonitor(dvt) as monitor:
        async for event in monitor:
            if isinstance(event, ConnectionDetectionEvent):
                logger.info(
                    f"Connection detected: {event.local_address.data.address}:{event.local_address.port} -> "
                    f"{event.remote_address.data.address}:{event.remote_address.port}"
                )


@cli.command("screenshot")
@async_command
async def dvt_screenshot(service_provider: ServiceProviderDep, out: Path) -> None:
    """Take device screenshot"""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        out.write_bytes(await Screenshot(dvt).get_screenshot())


@cli.command("xcuitest")
@async_command
async def xcuitest(
    service_provider: ServiceProviderDep,
    bundle_id: str,
    output_log: Annotated[
        Optional[Path],
        typer.Option(help="log debug output file location ( default to python logger )"),
    ] = None,
    timeout: Annotated[
        Optional[float],
        typer.Option(help="maximum execution time, in seconds"),
    ] = None,
    env: Annotated[
        Optional[list[str]],
        typer.Option(
            help="Environment variable to pass to process given as key=value (can be specified multiple times)"
        ),
    ] = None,
) -> None:
    """
    Start XCUITest

    \b
    Usage example:
    \b    python3 -m pymobiledevice3 developer dvt xcuitest com.facebook.WebDriverAgentRunner.xctrunner
    """
    s = XCUITestService(service_provider)
    consumer = await s.start(bundle_id, test_runner_env=dict(var.split("=", 1) for var in env or ()))

    async def stop_consumer_safely() -> None:
        try:
            await consumer.stop()
        except RuntimeError as e:
            if "no running event loop" not in str(e):
                raise
            logger.debug("Skipping consumer stop during event-loop teardown")

    if output_log is not None:

        class MyListener(XCUITestListener):
            def logDebugMessage_(self, args):
                output_log.write_text(args[0].value)

        consumer.set_listener(MyListener(consumer.main_dvt, consumer.main_chan, consumer.xctest_config))
    try:
        if timeout is not None:
            await asyncio.wait_for(consumer.consume(), timeout=timeout)
        else:
            await consumer.consume()
    except asyncio.TimeoutError:
        logger.warning("consumer task timed out, exiting")
        await stop_consumer_safely()
    except (asyncio.CancelledError, KeyboardInterrupt):
        logger.info("xcuitest interrupted, exiting")
        await stop_consumer_safely()
        raise


@cli.command("trace-codes")
@async_command
async def dvt_trace_codes(service_provider: ServiceProviderDep) -> None:
    """Print KDebug trace codes."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        device_info = DeviceInfo(dvt)
        print_json({hex(k): v for k, v in (await device_info.trace_codes()).items()})


@cli.command("name-for-uid")
@async_command
async def dvt_name_for_uid(service_provider: ServiceProviderDep, uid: int) -> None:
    """Print the assiciated username for the given uid."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        device_info = DeviceInfo(dvt)
        print(await device_info.name_for_uid(uid))


@cli.command("name-for-gid")
@async_command
async def dvt_name_for_gid(service_provider: ServiceProviderDep, gid: int) -> None:
    """Print the assiciated group name for the given gid."""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        device_info = DeviceInfo(dvt)
        print(await device_info.name_for_gid(gid))


@cli.command("oslog")
def dvt_oslog(service_provider: ServiceProviderDep, pid: int) -> None:
    """Sniff device oslog (not very stable, but includes more data and normal syslog)"""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt, ActivityTraceTap(dvt) as tap:
        for message in tap:
            message_pid = message.process
            # without message_type maybe signpost have event_type
            message_type = (
                message.message_type
                if hasattr(message, "message_type")
                else message.event_type
                if hasattr(message, "event_type")
                else "unknown"
            )
            sender_image_path = message.sender_image_path
            image_name = os.path.basename(sender_image_path)
            subsystem = message.subsystem
            category = message.category
            timestamp = datetime.now()

            if pid is not None and message_pid != pid:
                continue

            formatted_message = decode_message_format(message.message) if message.message else message.name

            if user_requested_colored_output():
                timestamp = typer.style(str(timestamp), bold=True)
                message_pid = typer.style(str(message_pid), "magenta")
                subsystem = typer.style(subsystem, "green")
                category = typer.style(category, "green")
                image_name = typer.style(image_name, "yellow")
                message_type = typer.style(message_type, "cyan")

            print(
                f"[{timestamp}][{subsystem}][{category}][{message_pid}][{image_name}] "
                f"<{message_type}>: {formatted_message}"
            )


@cli.command("energy")
@async_command
async def dvt_energy(service_provider: ServiceProviderDep, pid_list: list[str]) -> None:
    """Monitor the energy consumption for given PIDs"""

    if len(pid_list) == 0:
        logger.error("pid_list must not be empty")
        return

    pid_int_list = [int(pid) for pid in pid_list]

    async with (
        DvtSecureSocketProxyService(lockdown=service_provider) as dvt,
        EnergyMonitor(dvt, pid_int_list) as energy_monitor,
    ):
        async for telemetry in energy_monitor:
            logger.info(telemetry)


@cli.command("notifications")
def dvt_notifications(service_provider: ServiceProviderDep) -> None:
    """Monitor memory and app notifications"""
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt, Notifications(dvt) as notifications:
        for notification in notifications:
            logger.info(notification)


@cli.command("graphics")
@async_command
async def dvt_graphics(service_provider: ServiceProviderDep) -> None:
    """Monitor graphics-related information"""
    async with DvtSecureSocketProxyService(lockdown=service_provider) as dvt, Graphics(dvt) as graphics:
        async for stats in graphics:
            logger.info(stats)


@cli.command("har")
def dvt_har(service_provider: ServiceProviderDep) -> None:
    """
    Enable har-logging

    \b
    For more information, please read:
    \b    https://github.com/doronz88/harlogger?tab=readme-ov-file#enable-http-instrumentation-method
    """
    with DvtSecureSocketProxyService(lockdown=service_provider) as dvt:
        print("> Press Ctrl-C to abort")
        with ActivityTraceTap(dvt, enable_http_archive_logging=True) as tap:
            while True:
                tap.channel.receive_message()



import os
import uuid
from enum import Enum
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, Optional
from zipfile import ZIP_DEFLATED, BadZipFile, ZipFile

from parameter_decorators import str_to_path

from pymobiledevice3.exceptions import AppInstallError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.afc import AfcService
from pymobiledevice3.services.lockdown_service import LockdownService

GET_APPS_ADDITIONAL_INFO = {"ReturnAttributes": ["CFBundleIdentifier", "StaticDiskUsage", "DynamicDiskUsage"]}

TEMP_REMOTE_BASEDIR = "/PublicStaging/pymobiledevice3"


class ZipFileType(Enum):
    IPCC = "ipcc"
    IPA = "ipa"

    def is_ipcc(self) -> bool:
        return self == ZipFileType.IPCC

    def is_ipa(self) -> bool:
        return self == ZipFileType.IPA


def create_ipa_contents_from_directory(directory: str) -> bytes:
    payload_prefix = "Payload/" + os.path.basename(directory)
    with TemporaryDirectory() as temp_dir:
        zip_path = Path(temp_dir) / "ipa"
        with ZipFile(zip_path, "w", ZIP_DEFLATED) as zip_file:
            for root, _dirs, files in os.walk(directory):
                for file in files:
                    full_path = Path(root) / file
                    full_path.touch()
                    zip_file.write(full_path, arcname=f"{payload_prefix}/{os.path.relpath(full_path, directory)}")
        return zip_path.read_bytes()


def classify_zip_file(zip_bytes: bytes) -> ZipFileType:
    """checks the zipped bytes if it's a .ipcc or .ipa"""
    try:
        with ZipFile(BytesIO(zip_bytes), "r") as zip_file:
            dirs = next((name for name in zip_file.namelist() if "/" in name), "").split("/")

            if dirs[0] != "Payload":
                raise AppInstallError("package does not have a payload")
            if dirs[1].endswith(".app"):
                return ZipFileType.IPA
            elif dirs[1].endswith(".bundle"):
                return ZipFileType.IPCC
            else:
                raise AppInstallError("package does not have the appropriate folders structure")

    except BadZipFile as e:
        raise AppInstallError("Invalid bytes package") from e


class InstallationProxyService(LockdownService):
    SERVICE_NAME = "com.apple.mobile.installation_proxy"
    RSD_SERVICE_NAME = "com.apple.mobile.installation_proxy.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider):
        if isinstance(lockdown, LockdownClient):
            super().__init__(lockdown, self.SERVICE_NAME)
        else:
            super().__init__(lockdown, self.RSD_SERVICE_NAME)

    async def _watch_completion(self, handler: Optional[Callable] = None, *args) -> None:
        while True:
            response = await self.service.recv_plist()
            if not response:
                break
            error = response.get("Error")
            if error:
                raise AppInstallError(f"{error}: {response.get('ErrorDescription')}")
            completion = response.get("PercentComplete")
            if completion:
                if handler:
                    self.logger.debug("calling handler")
                    handler(completion, *args)
                self.logger.info(f"{completion}% Complete")
            if response.get("Status") == "Complete":
                self.logger.info("Installation succeed.")
                return
        raise AppInstallError()

    async def send_cmd_for_bundle_identifier(
        self,
        bundle_identifier: str,
        cmd: str = "Archive",
        options: Optional[dict] = None,
        handler: Optional[Callable] = None,
        *args,
    ) -> None:
        """send a low-level command to installation relay"""
        cmd: dict = {"Command": cmd, "ApplicationIdentifier": bundle_identifier}

        if options is None:
            options = {}

        cmd.update({"ClientOptions": options})
        await self.service.send_plist(cmd)
        await self._watch_completion(handler, *args)

    async def install(
        self, package_path: str, options: Optional[dict] = None, handler: Optional[Callable] = None, *args
    ) -> None:
        """install given ipa/ipcc from device path"""
        await self.install_from_local(package_path, "Install", options, handler, args)

    async def upgrade(
        self, ipa_path: str, options: Optional[dict] = None, handler: Optional[Callable] = None, *args
    ) -> None:
        """upgrade given ipa from device path"""
        await self.install_from_local(ipa_path, "Upgrade", options, handler, args)

    async def restore(
        self, bundle_identifier: str, options: Optional[dict] = None, handler: Optional[Callable] = None, *args
    ) -> None:
        """no longer supported on newer iOS versions"""
        await self.send_cmd_for_bundle_identifier(bundle_identifier, "Restore", options, handler, args)

    async def uninstall(
        self, bundle_identifier: str, options: Optional[dict] = None, handler: Optional[Callable] = None, *args
    ) -> None:
        """uninstall given bundle_identifier"""
        await self.send_cmd_for_bundle_identifier(bundle_identifier, "Uninstall", options, handler, args)

    async def install_from_bytes(
        self,
        package_bytes: bytes,
        cmd: str = "Install",
        options: Optional[dict] = None,
        handler: Optional[Callable] = None,
        *args,
    ) -> None:
        """upload given ipa/ipcc bytes object onto device and install it"""
        ipcc_mode = classify_zip_file(package_bytes).is_ipcc()

        if options is None:
            options = {}

        if ipcc_mode:
            options["PackageType"] = "CarrierBundle"

        async with AfcService(self.lockdown) as afc:
            fpath = f"{TEMP_REMOTE_BASEDIR}/{uuid.uuid4()}.{'ipcc' if ipcc_mode else 'ipa'}"
            try:
                if not ipcc_mode:
                    await afc.makedirs(TEMP_REMOTE_BASEDIR)
                    await afc.set_file_contents(fpath, package_bytes)
                else:
                    await self.upload_ipcc_from_bytes(package_bytes, fpath, afc)

                await self.send_package(cmd, options, handler, fpath, *args)
            finally:
                await afc.rm_single(fpath, force=True)

    @str_to_path("package_path")
    async def install_from_local(
        self,
        package_path: Path,
        cmd: str = "Install",
        options: Optional[dict] = None,
        handler: Optional[Callable] = None,
        developer: bool = False,
        *args,
    ) -> None:
        """upload given ipa/ipcc onto device and install it"""
        ipcc_mode = package_path.suffix == ".ipcc"

        if options is None:
            options = {}

        if ipcc_mode:
            options["PackageType"] = "CarrierBundle"
        else:
            if package_path.is_dir():
                # treat as app, convert into an ipa
                ipa_contents = create_ipa_contents_from_directory(str(package_path))
            else:
                # treat as ipa
                ipa_contents = package_path.read_bytes()

        if developer:
            options["PackageType"] = "Developer"

        async with AfcService(self.lockdown) as afc:
            fname = f"{TEMP_REMOTE_BASEDIR}/{uuid.uuid4()}.{'ipcc' if ipcc_mode else 'ipa'}"
            try:
                if not ipcc_mode:
                    await afc.makedirs(TEMP_REMOTE_BASEDIR)
                    await afc.set_file_contents(fname, ipa_contents)

                else:
                    await self.upload_ipcc_from_path(package_path, fname, afc)

                await self.send_package(cmd, options, handler, fname, *args)
            finally:
                await afc.rm_single(fname, force=True)

    async def send_package(self, cmd: str, options: Optional[dict], handler: Callable, package_path: str, *args):
        await self.service.send_plist({
            "Command": cmd,
            "ClientOptions": options,
            "PackagePath": package_path,
        })

        await self._watch_completion(handler, args)

    async def upload_ipcc_from_path(self, file: Path, remote_path: str, afc_client: AfcService) -> None:
        """Used to upload a .ipcc file to an iPhone as a folder"""
        with file.open("rb") as fb:
            file_name = file.name
            file_stream = BytesIO(fb.read())
            self.logger.info(f"Uploading {file_name} contents..")
            await self._upload_ipcc(file_stream, afc_client, remote_path)

    async def upload_ipcc_from_bytes(self, file_bytes: bytes, remote_path: str, afc_client: AfcService) -> None:
        """Used to upload a .ipcc bytes array to an iPhone as a folder"""
        file_stream = BytesIO(file_bytes)
        self.logger.info("Uploading IPCC from given bytes..")
        await self._upload_ipcc(file_stream, afc_client, remote_path)

    async def _upload_ipcc(self, file_stream: BytesIO, afc_client: AfcService, dst: str) -> None:
        self.logger.info(f"Uploading {dst} contents..")
        await afc_client.makedirs(dst)

        # we unpack it and upload it directly instead of saving it in a temp folder
        with ZipFile(file_stream, "r") as file_zip:
            for file_name in file_zip.namelist():
                if file_name.endswith(("/", "\\")):
                    await afc_client.makedirs(f"{dst}/{file_name}")
                    continue

                with file_zip.open(file_name) as inside_file_zip:
                    file_data = inside_file_zip.read()
                    await afc_client.makedirs(dst)
                    await afc_client.set_file_contents(f"{dst}/{file_name}", file_data)

        self.logger.info("Upload complete.")

    async def check_capabilities_match(
        self, capabilities: Optional[dict] = None, options: Optional[dict] = None
    ) -> dict:
        if options is None:
            options = {}
        cmd = {"Command": "CheckCapabilitiesMatch", "ClientOptions": options}

        if capabilities:
            cmd["Capabilities"] = capabilities

        await self.service.send_plist(cmd)
        return (await self.service.recv_plist()).get("LookupResult")

    async def browse(self, options: Optional[dict] = None, attributes: Optional[list[str]] = None) -> list[dict]:
        if options is None:
            options = {}
        if attributes:
            options["ReturnAttributes"] = attributes

        cmd = {"Command": "Browse", "ClientOptions": options}

        await self.service.send_plist(cmd)

        result = []
        while True:
            response = await self.service.recv_plist()
            if not response:
                break

            data = response.get("CurrentList")
            if data is not None:
                result += data

            if response.get("Status") == "Complete":
                break

        return result

    async def lookup(self, options: Optional[dict] = None) -> dict:
        """search installation database"""
        if options is None:
            options = {}
        cmd = {"Command": "Lookup", "ClientOptions": options}
        await self.service.send_plist(cmd)
        return (await self.service.recv_plist()).get("LookupResult")

    async def get_apps(
        self,
        application_type: str = "Any",
        calculate_sizes: bool = False,
        bundle_identifiers: Optional[list[str]] = None,
    ) -> dict[str, dict]:
        """get applications according to given criteria"""
        options = {}
        if bundle_identifiers is not None:
            options["BundleIDs"] = bundle_identifiers

        options["ApplicationType"] = application_type
        result = await self.lookup(options)
        if calculate_sizes:
            options.update(GET_APPS_ADDITIONAL_INFO)
            additional_info = await self.lookup(options)
            for bundle_identifier, app in additional_info.items():
                result[bundle_identifier].update(app)
        return result



import inspect
import socket
import sys
from datetime import datetime
from pathlib import Path

from pymobiledevice3.exceptions import FeatureNotSupportedError, OSNotSupportedError

DEFAULT_AFTER_IDLE_SEC = 3
DEFAULT_INTERVAL_SEC = 3
DEFAULT_MAX_FAILS = 3


def is_wsl() -> bool:
    try:
        with open("/proc/version") as f:
            version_info = f.read()
            return "Microsoft" in version_info or "WSL" in version_info
    except FileNotFoundError:
        return False


class OsUtils:
    _instance = None
    _os_name = None

    @classmethod
    def create(cls) -> "OsUtils":
        if cls._instance is None:
            cls._os_name = sys.platform
            if cls._os_name == "win32":
                from pymobiledevice3.osu.win_util import Win32

                cls._instance = Win32()
            elif cls._os_name == "darwin":
                from pymobiledevice3.osu.posix_util import Darwin

                cls._instance = Darwin()
            elif cls._os_name == "linux":
                from pymobiledevice3.osu.posix_util import Linux, Wsl

                cls._instance = Wsl() if is_wsl() else Linux()
            elif cls._os_name == "cygwin":
                from pymobiledevice3.osu.posix_util import Cygwin

                cls._instance = Cygwin()
            else:
                raise OSNotSupportedError(cls._os_name)
        return cls._instance

    @property
    def is_admin(self) -> bool:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def usbmux_address(self) -> tuple[str, int]:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def bonjour_timeout(self) -> int:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def loopback_header(self) -> bytes:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def access_denied_error(self) -> str:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    @property
    def pair_record_path(self) -> Path:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def get_ipv6_ips(self) -> list[str]:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def set_keepalive(
        self,
        sock: socket.socket,
        after_idle_sec: int = DEFAULT_AFTER_IDLE_SEC,
        interval_sec: int = DEFAULT_INTERVAL_SEC,
        max_fails: int = DEFAULT_MAX_FAILS,
    ) -> None:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def parse_timestamp(self, time_stamp) -> datetime:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def chown_to_non_sudo_if_needed(self, path: Path) -> None:
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def wait_return(self):
        raise FeatureNotSupportedError(self._os_name, inspect.currentframe().f_code.co_name)

    def get_homedir(self) -> Path:
        return Path.home()


def get_os_utils() -> OsUtils:
    return OsUtils.create()



from pymobiledevice3.services.remote_server import MessageAux


class Notifications:
    IDENTIFIER = "com.apple.instruments.server.services.mobilenotifications"

    def __init__(self, dvt):
        self._dvt = dvt
        self._channel = None

    async def __aenter__(self):
        self._channel = await self._dvt.make_channel(self.IDENTIFIER)
        await self._channel.setApplicationStateNotificationsEnabled_(MessageAux().append_obj(True))
        await self._channel.setMemoryNotificationsEnabled_(MessageAux().append_obj(True))
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._channel.setApplicationStateNotificationsEnabled_(MessageAux().append_obj(False))
        await self._channel.setMemoryNotificationsEnabled_(MessageAux().append_obj(False))

    async def __aiter__(self):
        while True:
            yield await self._dvt.recv_plist(self._channel)



class Alert:
    def __init__(self, session):
        """
        :param pymobiledevice3.services.web_protocol.automation_session.AutomationSession session: Automation session.
        """
        self.session = session

    async def accept(self):
        """Accepts the alert available."""
        await self.session.accept_current_javascript_dialog()

    async def dismiss(self):
        """Dismisses the alert available."""
        await self.session.dismiss_current_javascript_dialog()

    async def send_keys(self, text: str):
        """
        Send Keys to the Alert.
        :param text: Text to send to prompts.
        """
        await self.session.set_user_input_for_current_javascript_prompt(text)

    async def get_text(self) -> str:
        """Gets the text of the Alert."""
        return await self.session.message_of_current_javascript_dialog()



import dataclasses
import uuid
from datetime import datetime
from typing import Any

from construct import (
    Aligned,
    Array,
    Bytes,
    Const,
    CString,
    Default,
    Double,
    Enum,
    ExprAdapter,
    FlagsEnum,
    GreedyBytes,
    Hex,
    If,
    Int32ul,
    Int64sl,
    Int64ul,
    LazyBound,
    Pass,
    Prefixed,
    Probe,
    Struct,
    Switch,
    this,
)
from construct import Optional as ConstructOptional

XpcMessageType = Enum(
    Hex(Int32ul),
    NULL=0x00001000,
    BOOL=0x00002000,
    INT64=0x00003000,
    UINT64=0x00004000,
    DOUBLE=0x00005000,
    POINTER=0x00006000,
    DATE=0x00007000,
    DATA=0x00008000,
    STRING=0x00009000,
    UUID=0x0000A000,
    FD=0x0000B000,
    SHMEM=0x0000C000,
    MACH_SEND=0x0000D000,
    ARRAY=0x0000E000,
    DICTIONARY=0x0000F000,
    ERROR=0x00010000,
    CONNECTION=0x00011000,
    ENDPOINT=0x00012000,
    SERIALIZER=0x00013000,
    PIPE=0x00014000,
    MACH_RECV=0x00015000,
    BUNDLE=0x00016000,
    SERVICE=0x00017000,
    SERVICE_INSTANCE=0x00018000,
    ACTIVITY=0x00019000,
    FILE_TRANSFER=0x0001A000,
)
XpcFlags = FlagsEnum(
    Hex(Int32ul),
    ALWAYS_SET=0x00000001,
    PING=0x00000002,
    DATA_PRESENT=0x00000100,
    WANTING_REPLY=0x00010000,
    REPLY=0x00020000,
    FILE_TX_STREAM_REQUEST=0x00100000,
    FILE_TX_STREAM_RESPONSE=0x00200000,
    INIT_HANDSHAKE=0x00400000,
)
AlignedString = Aligned(4, CString("utf8"))
XpcNull = Pass
XpcBool = Int32ul
XpcInt64 = Int64sl
XpcUInt64 = Int64ul
XpcDouble = Double
XpcPointer = None
XpcDate = Int64ul
XpcData = Aligned(4, Prefixed(Int32ul, GreedyBytes))
XpcString = Aligned(4, Prefixed(Int32ul, CString("utf8")))
XpcUuid = Bytes(16)
XpcFd = Int32ul
XpcShmem = Struct("length" / Int32ul, Int32ul)
XpcArray = Prefixed(Int32ul, Struct("count" / Int32ul, "entries" / Array(this.count, LazyBound(lambda: XpcObject))))
XpcDictionaryEntry = Struct(
    "key" / AlignedString,
    "value" / LazyBound(lambda: XpcObject),
)
XpcDictionary = Prefixed(
    Int32ul,
    Struct(
        "count" / Hex(Int32ul),
        "entries" / If(this.count > 0, Array(this.count, XpcDictionaryEntry)),
    ),
)
XpcFileTransfer = Struct(
    "msg_id" / Int64ul,
    "data" / LazyBound(lambda: XpcObject),
)
XpcObject = Struct(
    "type" / XpcMessageType,
    "data"
    / Switch(
        this.type,
        {
            XpcMessageType.DICTIONARY: XpcDictionary,
            XpcMessageType.STRING: XpcString,
            XpcMessageType.INT64: XpcInt64,
            XpcMessageType.UINT64: XpcUInt64,
            XpcMessageType.DOUBLE: XpcDouble,
            XpcMessageType.BOOL: XpcBool,
            XpcMessageType.NULL: XpcNull,
            XpcMessageType.UUID: XpcUuid,
            XpcMessageType.POINTER: XpcPointer,
            XpcMessageType.DATE: XpcDate,
            XpcMessageType.DATA: XpcData,
            XpcMessageType.FD: XpcFd,
            XpcMessageType.SHMEM: XpcShmem,
            XpcMessageType.ARRAY: XpcArray,
            XpcMessageType.FILE_TRANSFER: XpcFileTransfer,
        },
        default=Probe(lookahead=1000),
    ),
)
XpcPayload = Struct(
    "magic" / Hex(Const(0x42133742, Int32ul)),
    "protocol_version" / Hex(Const(0x00000005, Int32ul)),
    "obj" / XpcObject,
)
XpcWrapper = Struct(
    "magic" / Hex(Const(0x29B00B92, Int32ul)),
    "flags" / Default(XpcFlags, XpcFlags.ALWAYS_SET),
    "message"
    / Prefixed(
        ExprAdapter(Int64ul, lambda obj, context: obj + 8, lambda obj, context: obj - 8),
        Struct(
            "message_id" / Hex(Default(Int64ul, 0)),
            "payload" / ConstructOptional(XpcPayload),
        ),
    ),
)


class XpcInt64Type(int):
    pass


class XpcUInt64Type(int):
    pass


@dataclasses.dataclass
class FileTransferType:
    transfer_size: int


def _decode_xpc_dictionary(xpc_object) -> dict:
    if xpc_object.data.count == 0:
        return {}
    result = {}
    for entry in xpc_object.data.entries:
        result[entry.key] = decode_xpc_object(entry.value)
    return result


def _decode_xpc_array(xpc_object) -> list:
    result = []
    for entry in xpc_object.data.entries:
        result.append(decode_xpc_object(entry))
    return result


def _decode_xpc_bool(xpc_object) -> bool:
    return bool(xpc_object.data)


def _decode_xpc_int64(xpc_object) -> XpcInt64Type:
    return XpcInt64Type(xpc_object.data)


def _decode_xpc_uint64(xpc_object) -> XpcUInt64Type:
    return XpcUInt64Type(xpc_object.data)


def _decode_xpc_uuid(xpc_object) -> uuid.UUID:
    return uuid.UUID(bytes=xpc_object.data)


def _decode_xpc_string(xpc_object) -> str:
    return xpc_object.data


def _decode_xpc_data(xpc_object) -> bytes:
    return xpc_object.data


def _decode_xpc_date(xpc_object) -> datetime:
    # Convert from nanoseconds to seconds
    return datetime.fromtimestamp(xpc_object.data / 1000000000)


def _decode_xpc_file_transfer(xpc_object) -> FileTransferType:
    return FileTransferType(transfer_size=_decode_xpc_dictionary(xpc_object.data.data)["s"])


def _decode_xpc_double(xpc_object) -> float:
    return xpc_object.data


def _decode_xpc_null(xpc_object) -> None:
    return None


def decode_xpc_object(xpc_object) -> Any:
    decoders = {
        XpcMessageType.DICTIONARY: _decode_xpc_dictionary,
        XpcMessageType.ARRAY: _decode_xpc_array,
        XpcMessageType.BOOL: _decode_xpc_bool,
        XpcMessageType.INT64: _decode_xpc_int64,
        XpcMessageType.UINT64: _decode_xpc_uint64,
        XpcMessageType.UUID: _decode_xpc_uuid,
        XpcMessageType.STRING: _decode_xpc_string,
        XpcMessageType.DATA: _decode_xpc_data,
        XpcMessageType.DATE: _decode_xpc_date,
        XpcMessageType.FILE_TRANSFER: _decode_xpc_file_transfer,
        XpcMessageType.DOUBLE: _decode_xpc_double,
        XpcMessageType.NULL: _decode_xpc_null,
    }
    decoder = decoders.get(xpc_object.type)
    if decoder is None:
        raise TypeError(f"deserialize error: {xpc_object}")
    return decoder(xpc_object)


def _build_xpc_array(payload: list) -> dict:
    entries = []
    for entry in payload:
        entry = _build_xpc_object(entry)
        entries.append(entry)
    return {"type": XpcMessageType.ARRAY, "data": {"count": len(entries), "entries": entries}}


def _build_xpc_dictionary(payload: dict) -> dict:
    entries = []
    for key, value in payload.items():
        entry = {"key": key, "value": _build_xpc_object(value)}
        entries.append(entry)
    return {
        "type": XpcMessageType.DICTIONARY,
        "data": {
            "count": len(entries),
            "entries": entries,
        },
    }


def _build_xpc_bool(payload: bool) -> dict:
    return {
        "type": XpcMessageType.BOOL,
        "data": payload,
    }


def _build_xpc_string(payload: str) -> dict:
    return {
        "type": XpcMessageType.STRING,
        "data": payload,
    }


def _build_xpc_data(payload: bool) -> dict:
    return {
        "type": XpcMessageType.DATA,
        "data": payload,
    }


def _build_xpc_double(payload: float) -> dict:
    return {
        "type": XpcMessageType.DOUBLE,
        "data": payload,
    }


def _build_xpc_uuid(payload: uuid.UUID) -> dict:
    return {
        "type": XpcMessageType.UUID,
        "data": payload.bytes,
    }


def _build_xpc_null(payload: None) -> dict:
    return {
        "type": XpcMessageType.NULL,
        "data": None,
    }


def _build_xpc_uint64(payload: XpcUInt64Type) -> dict:
    return {
        "type": XpcMessageType.UINT64,
        "data": payload,
    }


def _build_xpc_int64(payload: XpcInt64Type) -> dict:
    return {
        "type": XpcMessageType.INT64,
        "data": payload,
    }


def _build_xpc_object(payload: Any) -> dict:
    if payload is None:
        return _build_xpc_null(payload)
    payload_builders = {
        list: _build_xpc_array,
        dict: _build_xpc_dictionary,
        bool: _build_xpc_bool,
        str: _build_xpc_string,
        bytes: _build_xpc_data,
        bytearray: _build_xpc_data,
        float: _build_xpc_double,
        uuid.UUID: _build_xpc_uuid,
        "XpcUInt64Type": _build_xpc_uint64,
        "XpcInt64Type": _build_xpc_int64,
    }
    builder = payload_builders.get(type(payload), payload_builders.get(type(payload).__name__))
    if builder is None:
        raise TypeError(f"unrecognized type for: {payload} {type(payload)}")
    return builder(payload)


def create_xpc_wrapper(d: dict, message_id: int = 0, wanting_reply: bool = False) -> bytes:
    flags = XpcFlags.ALWAYS_SET
    if len(d.keys()) > 0:
        flags |= XpcFlags.DATA_PRESENT
    if wanting_reply:
        flags |= XpcFlags.WANTING_REPLY

    xpc_payload = {"message_id": message_id, "payload": {"obj": _build_xpc_object(d)}}

    xpc_wrapper = {"flags": flags, "message": xpc_payload}
    return XpcWrapper.build(xpc_wrapper)



import asyncio
import json
import typing
from collections.abc import AsyncGenerator
from dataclasses import dataclass
from enum import Enum, IntEnum

from packaging.version import Version

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.lockdown_service_provider import LockdownServiceProvider
from pymobiledevice3.services.remote_server import MessageAux, RemoteServer


class SerializedObject:
    def __init__(self, fields: dict):
        self._fields = fields


class AXAuditInspectorFocus_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)

    @property
    def caption(self) -> str:
        return self._fields.get("CaptionTextValue_v1")

    @property
    def spoken_description(self) -> str:
        return self._fields.get("SpokenDescriptionValue_v1")

    @property
    def element(self) -> bytes:
        return self._fields.get("ElementValue_v1")

    @property
    def platform_identifier(self) -> str:
        """Converts the element bytes to a hexadecimal string."""
        return self.element.identifier.hex().upper()

    @property
    def estimated_uid(self) -> str:
        """Generates a UID from the platform identifier."""
        hex_value = self.platform_identifier

        if len(hex_value) % 2 != 0:
            raise ValueError("Hex value length must be even.")

        hex_bytes = bytes.fromhex(hex_value)

        if len(hex_bytes) < 16:
            raise ValueError("Hex value must contain at least 16 bytes.")

        # Extract TimeLow bytes (indexes 12 to 15)
        time_low_bytes = hex_bytes[12:16]
        time_low = time_low_bytes.hex().upper()

        # Extract ClockSeq bytes (indexes 0 to 1)
        clock_seq_bytes = hex_bytes[0:2]
        clock_seq = clock_seq_bytes.hex().upper()

        # Construct UID with placeholder values for unused parts
        uid = f"{time_low}-0000-0000-{clock_seq}-000000000000"

        return uid

    def to_dict(self) -> dict:
        """Serializes the focus element into a dictionary."""
        return {
            "platform_identifier": self.platform_identifier,
            "estimated_uid": self.estimated_uid,
            "caption": self.caption,
            "spoken_description": self.spoken_description,
        }

    def __str__(self):
        return f"<Focused ElementCaption: {self.caption}>"


class AXAuditElement_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)

    @property
    def identifier(self) -> bytes:
        return self._fields["PlatformElementValue_v1"]

    def __repr__(self):
        return f"<Element: {self.identifier}>"


class AXAuditInspectorSection_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)


class AXAuditElementAttribute_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)


class AXAuditDeviceSetting_v1(SerializedObject):
    FIELDS = ("IdentiifierValue_v1", "CurrentValueNumber_v1")

    def __init__(self, fields):
        super().__init__(fields)
        for k in self.FIELDS:
            if k not in self._fields:
                self._fields[k] = None

    @property
    def key(self) -> str:
        return self._fields["IdentiifierValue_v1"]

    @property
    def value(self) -> typing.Any:
        return self._fields["CurrentValueNumber_v1"]

    def __str__(self) -> str:
        return f"<AXAuditDeviceSetting_v1 {self.key} = {self.value}>"


class AuditType(IntEnum):
    DYNAMIC_TEXT = 3001
    DYNAMIC_TEXT_ALT = 3002
    TEXT_CLIPPED = 3003
    ELEMENT_DETECTION = 1000
    SUFFICIENT_ELEMENT_DESCRIPTION = 5000
    HIT_REGION = 100
    CONTRAST = 12
    CONTRAST_ALT = 13


AUDIT_TYPE_DESCRIPTIONS = {
    AuditType.DYNAMIC_TEXT: "testTypeDynamicText",
    AuditType.DYNAMIC_TEXT_ALT: "testTypeDynamicText",
    AuditType.TEXT_CLIPPED: "testTypeTextClipped",
    AuditType.ELEMENT_DETECTION: "testTypeElementDetection",
    AuditType.SUFFICIENT_ELEMENT_DESCRIPTION: "testTypeSufficientElementDescription",
    AuditType.HIT_REGION: "testTypeHitRegion",
    AuditType.CONTRAST: "testTypeContrast",
    AuditType.CONTRAST_ALT: "testTypeContrast",
}


class AXAuditIssue_v1(SerializedObject):
    FIELDS = (
        "ElementRectValue_v1",
        "IssueClassificationValue_v1",
        "FontSizeValue_v1",
        "MLGeneratedDescriptionValue_v1",
        "ElementLongDescExtraInfo_v1",
        "BackgroundColorValue_v1",
        "ForegroundColorValue_v1",
    )

    def __init__(self, fields):
        super().__init__(fields)

        for k in self.FIELDS:
            if k not in self._fields:
                self._fields[k] = None

    @property
    def rect(self) -> str:
        return self._fields["ElementRectValue_v1"]

    @property
    def issue_type(self) -> typing.Any:
        issue_classification = self._fields["IssueClassificationValue_v1"]
        if issue_classification in AUDIT_TYPE_DESCRIPTIONS:
            return AUDIT_TYPE_DESCRIPTIONS[AuditType(issue_classification)]
        else:
            return issue_classification

    @property
    def ml_generated_description(self) -> typing.Any:
        return self._fields["MLGeneratedDescriptionValue_v1"]

    @property
    def long_description_extra_info(self) -> typing.Any:
        return self._fields["ElementLongDescExtraInfo_v1"]

    @property
    def font_size(self) -> typing.Any:
        return self._fields["FontSizeValue_v1"]

    @property
    def foreground_color(self) -> typing.Any:
        return self._fields["ForegroundColorValue_v1"]

    @property
    def background_color(self) -> typing.Any:
        return self._fields["BackgroundColorValue_v1"]

    def json(self) -> dict:
        resp = {
            "element_rect_value": self.rect,
            "issue_classification": self.issue_type,
            "font_size": self.font_size,
            "ml_generated_description": self.ml_generated_description,
            "long_description_extra_info": self.long_description_extra_info,
        }
        # Include foreground and background colors when issue type is 'testTypeContrast'
        if self._fields["IssueClassificationValue_v1"] in {AuditType.CONTRAST, AuditType.CONTRAST_ALT}:
            resp["foreground_color"] = self.foreground_color
            resp["background_color"] = self.background_color
        return resp

    def __str__(self) -> str:
        return json.dumps(self.json())


SERIALIZABLE_OBJECTS = {
    "AXAuditDeviceSetting_v1": AXAuditDeviceSetting_v1,
    "AXAuditInspectorFocus_v1": AXAuditInspectorFocus_v1,
    "AXAuditElement_v1": AXAuditElement_v1,
    "AXAuditInspectorSection_v1": AXAuditInspectorSection_v1,
    "AXAuditElementAttribute_v1": AXAuditElementAttribute_v1,
    "AXAuditIssue_v1": AXAuditIssue_v1,
}


@dataclass
class Event:
    name: str
    data: SerializedObject


class Direction(Enum):
    Previous = 3
    Next = 4
    First = 5
    Last = 6


def deserialize_object(d):
    if not isinstance(d, dict):
        if isinstance(d, list):
            return [deserialize_object(x) for x in d]
        return d

    if "ObjectType" not in d:
        # simple dictionary
        new_dict = {}
        for k, v in d.items():
            new_dict[k] = deserialize_object(v)
        return new_dict

    if d["ObjectType"] == "passthrough":
        return deserialize_object(d["Value"])
    else:
        return SERIALIZABLE_OBJECTS[d["ObjectType"]](deserialize_object(d["Value"]))


class AccessibilityAudit(RemoteServer):
    SERVICE_NAME = "com.apple.accessibility.axAuditDaemon.remoteserver"
    RSD_SERVICE_NAME = "com.apple.accessibility.axAuditDaemon.remoteserver.shim.remote"

    def __init__(self, lockdown: LockdownServiceProvider):
        service_name = self.SERVICE_NAME if isinstance(lockdown, LockdownClient) else self.RSD_SERVICE_NAME

        super().__init__(
            lockdown,
            service_name,
            remove_ssl_context=True,
            is_developer_service=False,
        )

        self.product_version = Version(lockdown.product_version)
        self._connected = False
        self._initial_messages_to_flush = 2 if self.product_version >= Version("15.0") else 1
        self._initial_messages_flushed = False

    async def _ensure_ready(self) -> None:
        if self._connected:
            return
        await self.connect()
        if not self._initial_messages_flushed:
            for _ in range(self._initial_messages_to_flush):
                try:
                    await asyncio.wait_for(self.recv_plist(), timeout=0.3)
                except asyncio.TimeoutError:
                    break
            self._initial_messages_flushed = True
        self._connected = True

    async def capabilities(self) -> list[str]:
        await self._ensure_ready()
        await self.broadcast.deviceCapabilities()
        return (await self.recv_plist())[0]

    async def run_audit(self, value: list) -> list[AXAuditIssue_v1]:
        await self._ensure_ready()
        if self.product_version >= Version("15.0"):
            await self.broadcast.deviceBeginAuditTypes_(MessageAux().append_obj(value))
        else:
            await self.broadcast.deviceBeginAuditCaseIDs_(MessageAux().append_obj(value))

        while True:
            message = await self.recv_plist()
            if message[1] is None or message[0] != "hostDeviceDidCompleteAuditCategoriesWithAuditIssues:":
                continue
            return deserialize_object(message[1])[0]["value"]

    async def supported_audits_types(self) -> list:
        await self._ensure_ready()
        if self.product_version >= Version("15.0"):
            await self.broadcast.deviceAllSupportedAuditTypes()
        else:
            await self.broadcast.deviceAllAuditCaseIDs()
        return deserialize_object((await self.recv_plist())[0])

    async def settings(self) -> list[AXAuditDeviceSetting_v1]:
        await self._ensure_ready()
        await self.broadcast.deviceAccessibilitySettings()
        return deserialize_object((await self.recv_plist())[0])

    async def perform_handshake(self) -> None:
        # this service acts differently from others, requiring no handshake
        pass

    async def set_app_monitoring_enabled(self, value: bool) -> None:
        await self._ensure_ready()
        await self.broadcast.deviceSetAppMonitoringEnabled_(MessageAux().append_obj(value), expects_reply=False)

    async def set_monitored_event_type(self, event_type: typing.Optional[int] = None) -> None:
        await self._ensure_ready()
        if event_type is None:
            event_type = 0
        await self.broadcast.deviceInspectorSetMonitoredEventType_(
            MessageAux().append_obj(event_type), expects_reply=False
        )

    async def set_show_ignored_elements(self, value: bool) -> None:
        await self._ensure_ready()
        await self.broadcast.deviceInspectorShowIgnoredElements_(
            MessageAux().append_obj(int(value)), expects_reply=False
        )

    async def set_show_visuals(self, value: bool) -> None:
        await self._ensure_ready()
        await self.broadcast.deviceInspectorShowVisuals_(MessageAux().append_obj(int(value)), expects_reply=False)

    async def iter_events(
        self, app_monitoring_enabled=True, monitored_event_type: typing.Optional[int] = None
    ) -> AsyncGenerator[Event, None]:
        await self._ensure_ready()
        await self.set_app_monitoring_enabled(app_monitoring_enabled)
        await self.set_monitored_event_type(monitored_event_type)

        while True:
            message = await self.recv_plist()
            if message[1] is None:
                continue
            data = [x["value"] for x in message[1]]
            yield Event(name=message[0], data=deserialize_object(data))

    async def move_focus_next(self) -> None:
        await self.move_focus(Direction.Next)

    async def perform_press(self, element: bytes) -> None:
        """simulate click (can be used only for processes with task_for_pid-allow"""
        await self._ensure_ready()
        element = {
            "ObjectType": "AXAuditElement_v1",
            "Value": {
                "ObjectType": "passthrough",
                "Value": {
                    "PlatformElementValue_v1": {"ObjectType": "passthrough"},
                    "Value": element,
                },
            },
        }

        action = {
            "ObjectType": "AXAuditElementAttribute_v1",
            "Value": {
                "ObjectType": "passthrough",
                "Value": {
                    "AttributeNameValue_v1": {
                        "ObjectType": "passthrough",
                        "Value": "AXAction-2010",
                    },
                    "DisplayAsTree_v1": {
                        "ObjectType": "passthrough",
                        "Value": 0,
                    },
                    "HumanReadableNameValue_v1": {
                        "ObjectType": "passthrough",
                        "Value": "Activate",
                    },
                    "IsInternal_v1": {
                        "ObjectType": "passthrough",
                        "Value": 0,
                    },
                    "PerformsActionValue_v1": {
                        "ObjectType": "passthrough",
                        "Value": 1,
                    },
                    "SettableValue_v1": {
                        "ObjectType": "passthrough",
                        "Value": 0,
                    },
                    "ValueTypeValue_v1": {
                        "ObjectType": "passthrough",
                        "Value": 1,
                    },
                },
            },
        }

        await self.broadcast.deviceElement_performAction_withValue_(
            MessageAux().append_obj(element).append_obj(action).append_obj(0), expects_reply=False
        )

    async def move_focus(self, direction: Direction) -> None:
        await self._ensure_ready()
        options = {
            "ObjectType": "passthrough",
            "Value": {
                "allowNonAX": {
                    "ObjectType": "passthrough",
                    "Value": 0,
                },
                "direction": {
                    "ObjectType": "passthrough",
                    "Value": direction.value,
                },
                "includeContainers": {
                    "ObjectType": "passthrough",
                    "Value": 1,
                },
            },
        }

        await self.broadcast.deviceInspectorMoveWithOptions_(MessageAux().append_obj(options), expects_reply=False)

    async def set_setting(self, name: str, value: typing.Any) -> None:
        await self._ensure_ready()
        setting = {
            "ObjectType": "AXAuditDeviceSetting_v1",
            "Value": {
                "ObjectType": "passthrough",
                "Value": {
                    "CurrentValueNumber_v1": {"ObjectType": "passthrough", "Value": True},
                    "EnabledValue_v1": {"ObjectType": "passthrough", "Value": True},
                    "IdentiifierValue_v1": {"ObjectType": "passthrough", "Value": name},
                    "SettingTypeValue_v1": {"ObjectType": "passthrough", "Value": 3},
                    "SliderTickMarksValue_v1": {"ObjectType": "passthrough", "Value": 0},
                },
            },
        }
        await self.broadcast.deviceUpdateAccessibilitySetting_withValue_(
            MessageAux().append_obj(setting).append_obj({"ObjectType": "passthrough", "Value": value}),
            expects_reply=False,
        )

    async def reset_settings(self) -> None:
        await self._ensure_ready()
        await self.broadcast.deviceResetToDefaultAccessibilitySettings()

    async def iter_elements(self) -> AsyncGenerator[AXAuditInspectorFocus_v1, None]:
        await self._ensure_ready()
        await self.set_app_monitoring_enabled(True)
        await self.set_monitored_event_type()

        # Every focus change is expected to publish "hostInspectorCurrentElementChanged:".
        await self.move_focus_next()
        visited_identifiers = set()
        consecutive_timeouts = 0

        while True:
            try:
                message = await asyncio.wait_for(self.recv_plist(), timeout=1.0)
                consecutive_timeouts = 0
            except asyncio.TimeoutError as err:
                consecutive_timeouts += 1
                if consecutive_timeouts >= 5:
                    raise TimeoutError("timed out waiting for accessibility focus events") from err
                await self.move_focus_next()
                continue
            if message[1] is None:
                continue
            data = [x["value"] for x in message[1]]
            event = Event(name=message[0], data=deserialize_object(data))
            if event.name != "hostInspectorCurrentElementChanged:":
                # ignore any other events
                continue

            # each such event should contain exactly one element that became in focus
            current_item = event.data[0]
            current_identifier = current_item.platform_identifier

            if current_identifier in visited_identifiers:
                break  # Exit if we've seen this element before (loop detected)

            yield current_item
            visited_identifiers.add(current_identifier)
            await self.move_focus_next()



import ast
import logging

from typer_injector import InjectingTyper

from pymobiledevice3.cli.cli_common import OSUTILS, ServiceProviderDep, async_command
from pymobiledevice3.services.accessibilityaudit import AccessibilityAudit

logger = logging.getLogger(__name__)


cli = InjectingTyper(
    name="settings",
    help="accessibility settings",
    no_args_is_help=True,
)


@cli.command("show")
@async_command
async def accessibility_settings_show(service_provider: ServiceProviderDep) -> None:
    """show current settings"""
    for setting in await AccessibilityAudit(service_provider).settings():
        print(setting)


@cli.command("set")
@async_command
async def accessibility_settings_set(service_provider: ServiceProviderDep, setting: str, value: str) -> None:
    """
    change current settings

    in order to list all available use the "show" command
    """
    service = AccessibilityAudit(service_provider)
    await service.set_setting(setting, ast.literal_eval(value))
    OSUTILS.wait_return()


@cli.command("reset")
@async_command
async def accessibility_settings_reset(service_provider: ServiceProviderDep) -> None:
    """
    reset accessibility settings to default
    """
    service = AccessibilityAudit(service_provider)
    await service.reset_settings()
